import numpy as np
from copy import deepcopy
import torch

from ..constants import ets as constants
from ..constants.constants import PredictionKeys
from ..estimators.stan_estimator import StanEstimatorMCMC, StanEstimatorMAP, StanEstimatorVI
from ..exceptions import IllegalArgument
from ..initializer.ets import ETSInitializer
from .template import BaseTemplate, FullBayesianTemplate, AggregatedPosteriorTemplate, MAPTemplate


class BaseETS(BaseTemplate):
    """

    Parameters
    ----------
    seasonality : int
        Length of seasonality
    seasonality_sm_input : float
        float value between [0, 1], applicable only if `seasonality` > 1. A larger value puts
        more weight on the current seasonality.
        If None, the model will estimate this value.
    level_sm_input : float
        float value between [0.0001, 1]. A larger value puts more weight on the current level.
        If None, the model will estimate this value.

    Other Parameters
    ----------------
    **kwargs: additional arguments passed into orbit.estimators.stan_estimator or orbit.estimators.pyro_estimator

    """
    # data labels for sampler
    _data_input_mapper = constants.DataInputMapper
    # used to match name of `*.stan` or `*.pyro` file to look for the model
    _model_name = 'ets'

    def __init__(self, seasonality=None, seasonality_sm_input=None, level_sm_input=None, **kwargs):
        super().__init__(**kwargs)  # create estimator in base class
        self.seasonality = seasonality

        # fixed smoothing parameters config
        self.seasonality_sm_input = seasonality_sm_input
        self.level_sm_input = level_sm_input

        # set private var to arg value
        # if None set default in _set_default_args()
        self._seasonality = self.seasonality
        self._seasonality_sm_input = self.seasonality_sm_input
        self._level_sm_input = self.level_sm_input

    def _set_static_attributes(self):
        """Override function from Base Template"""
        # setting defaults and proper data type
        if self.seasonality_sm_input is None:
            self._seasonality_sm_input = -1
        if self.level_sm_input is None:
            self._level_sm_input = -1
        elif self.level_sm_input < 0.0001 or self.level_sm_input > 1:
            raise IllegalArgument('only values between [0.0001, 1] are supported for level_sm_input '
                                  'to build a model with meaningful trend.')
        if self.seasonality is None:
            self._seasonality = -1

    def _set_init_values(self):
        """Override function from Base Template
        """
        # init_values_partial = partial(init_values_callable, seasonality=seasonality)
        # partialfunc does not work when passed to PyStan because PyStan uses
        # inspect.getargspec(func) which seems to raise an exception with keyword-only args
        # caused by using partialfunc
        # lambda does not work in serialization in pickle
        # callable object as an alternative workaround
        if self._seasonality > 1:
            init_values_callable = ETSInitializer(
                self._seasonality
            )
            self._init_values = init_values_callable

    def _set_model_param_names(self):
        """Set posteriors keys to extract from sampling/optimization api"""
        self._model_param_names += [param.value for param in constants.BaseSamplingParameters]

        # append seasonality param names
        if self._seasonality > 1:
            self._model_param_names += [param.value for param in constants.SeasonalitySamplingParameters]

    def _predict(self, posterior_estimates, df, include_error=False, **kwargs):
        """Vectorized version of prediction math"""
        ################################################################
        # Prediction Attributes
        ################################################################
        n_forecast_steps = self.prediction_input_meta['n_forecast_steps']
        start = self.prediction_input_meta['start']
        trained_len = self.num_of_observations
        full_len = trained_len + n_forecast_steps

        ################################################################
        # Model Attributes
        ################################################################
        model = deepcopy(posterior_estimates)
        for k, v in model.items():
            model[k] = torch.from_numpy(v)

        # We can pull any arbitrary value from teh dictionary because we hold the
        # safe assumption: the length of the first dimension is always the number of samples
        # thus can be safely used to determine `num_sample`. If predict_method is anything
        # other than full, the value here should be 1
        arbitrary_posterior_value = list(model.values())[0]
        num_sample = arbitrary_posterior_value.shape[0]

        # seasonality components
        seasonality_levels = model.get(
            constants.SeasonalitySamplingParameters.SEASONALITY_LEVELS.value)
        seasonality_smoothing_factor = model.get(
            constants.SeasonalitySamplingParameters.SEASONALITY_SMOOTHING_FACTOR.value
        )

        # trend components
        level_smoothing_factor = model.get(
            constants.BaseSamplingParameters.LEVEL_SMOOTHING_FACTOR.value)
        local_trend_levels = model.get(constants.BaseSamplingParameters.LOCAL_TREND_LEVELS.value)
        residual_sigma = model.get(constants.BaseSamplingParameters.RESIDUAL_SIGMA.value)

        ################################################################
        # Seasonality Component
        ################################################################

        # calculate seasonality component
        if self._seasonality > 1:
            if full_len <= seasonality_levels.shape[1]:
                seasonal_component = seasonality_levels[:, :full_len]
            else:
                seasonality_forecast_length = full_len - seasonality_levels.shape[1]
                seasonality_forecast_matrix \
                    = torch.zeros((num_sample, seasonality_forecast_length), dtype=torch.double)
                seasonal_component = torch.cat(
                    (seasonality_levels, seasonality_forecast_matrix), dim=1)
        else:
            seasonal_component = torch.zeros((num_sample, full_len), dtype=torch.double)

        ################################################################
        # Trend Component
        ################################################################

        # calculate level component.
        # However, if predicted end of period > training period, update with out-of-samples forecast
        if full_len <= trained_len:
            trend_component = local_trend_levels[:, :full_len]
            # in-sample error are iids
            if include_error:
                error_value = np.random.normal(
                    loc=0,
                    scale=residual_sigma.unsqueeze(-1),
                    size=trend_component.shape
                )

                error_value = torch.from_numpy(error_value).double()
                trend_component += error_value
        else:
            trend_component = local_trend_levels
            # in-sample error are iids
            if include_error:
                error_value = np.random.normal(
                    loc=0,
                    scale=residual_sigma.unsqueeze(-1),
                    size=trend_component.shape
                )

                error_value = torch.from_numpy(error_value).double()
                trend_component += error_value

            trend_forecast_matrix = torch.zeros((num_sample, n_forecast_steps), dtype=torch.double)
            trend_component = torch.cat((trend_component, trend_forecast_matrix), dim=1)

            last_local_trend_level = local_trend_levels[:, -1]

            for idx in range(trained_len, full_len):
                trend_component[:, idx] = last_local_trend_level

                if include_error:
                    error_value = np.random.normal(
                        scale=residual_sigma,
                        size=num_sample
                    )
                    error_value = torch.from_numpy(error_value).double()
                    trend_component[:, idx] += error_value

                new_local_trend_level = \
                    level_smoothing_factor * trend_component[:, idx] \
                    + (1 - level_smoothing_factor) * last_local_trend_level

                if self._seasonality > 1 and idx + self._seasonality < full_len:
                    seasonal_component[:, idx + self._seasonality] = \
                        seasonality_smoothing_factor.flatten() \
                        * (trend_component[:, idx] + seasonal_component[:, idx] -
                           new_local_trend_level) \
                        + (1 - seasonality_smoothing_factor.flatten()) * seasonal_component[:, idx]

                last_local_trend_level = new_local_trend_level

        ################################################################
        # Combine Components
        ################################################################

        # trim component with right start index
        trend_component = trend_component[:, start:]
        seasonal_component = seasonal_component[:, start:]

        # sum components
        pred_array = trend_component + seasonal_component

        pred_array = pred_array.numpy()
        trend_component = trend_component.numpy()
        seasonal_component = seasonal_component.numpy()

        out = {
            PredictionKeys.PREDICTION.value: pred_array,
            PredictionKeys.TREND.value: trend_component,
            PredictionKeys.SEASONALITY.value: seasonal_component
        }

        return out


class ETSMAP(MAPTemplate, BaseETS):
    """Concrete ETS model for MAP (Maximum a Posteriori) prediction

    Similar to `ETSAggregated` but prediction is based on Maximum a Posteriori (aka Mode)
    of the posterior.

    This model only supports MAP estimating `estimator_type`s

    """
    _supported_estimator_types = [StanEstimatorMAP]

    def __init__(self, estimator_type=StanEstimatorMAP, **kwargs):
        super().__init__(estimator_type=estimator_type, **kwargs)


class ETSFull(FullBayesianTemplate, BaseETS):
    """Concrete ETS model for full Bayesian prediction"""
    _supported_estimator_types = [StanEstimatorMCMC, StanEstimatorVI]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class ETSAggregated(AggregatedPosteriorTemplate, BaseETS):
    """Concrete ETS model for aggregated posterior prediction"""
    _supported_estimator_types = [StanEstimatorMCMC, StanEstimatorVI]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)



import pandas as pd
import numpy as np
import math
from scipy.stats import nct
import torch
from copy import deepcopy
import matplotlib.pyplot as plt

from ..constants import ktrx as constants
from ..constants.constants import (
    CoefPriorDictKeys,
)
from ..constants.ktrx import (
    DEFAULT_REGRESSOR_SIGN,
    DEFAULT_COEFFICIENTS_INIT_KNOT_SCALE,
    DEFAULT_COEFFICIENTS_INIT_KNOT_LOC,
    DEFAULT_COEFFICIENTS_KNOT_SCALE,
    DEFAULT_LOWER_BOUND_SCALE_MULTIPLIER,
    DEFAULT_UPPER_BOUND_SCALE_MULTIPLIER,
)

from ..estimators.pyro_estimator import PyroEstimatorVI
from ..exceptions import IllegalArgument, ModelException, PredictionException
from ..utils.general import is_ordered_datetime
from ..utils.kernels import gauss_kernel, sandwich_kernel
from ..utils.features import make_fourier_series_df
from .template import BaseTemplate, FullBayesianTemplate, AggregatedPosteriorTemplate


class BaseKTRX(BaseTemplate):
    """Base KTRX model object with shared functionality for PyroVI method

    Parameters
    ----------
    level_knot_scale : float
        sigma for level; default to be .1
    regressor_col : array-like strings
        regressor columns
    regressor_sign : list
        list of signs with '=' for regular regressor and '+' for positive regressor
    regressor_init_knot_loc : list
        list of regressor knot pooling mean priors, default to be 0's
    regressor_init_knot_scale : list
        list of regressor knot pooling sigma's to control the pooling strength towards the grand mean of regressors;
        default to be 1.
    regressor_knot_scale : list
        list of regressor knot sigma priors; default to be 0.1.
    span_coefficients : float between (0, 1)
        window width to decide the number of windows for the regression term
    rho_coefficients : float
        sigma in the Gaussian kernel for the regression term
    degree of freedom : int
        degree of freedom for error t-distribution
    coef_prior_list : list of dicts
        each dict in the list should have keys as
        'name', prior_start_tp_idx' (inclusive), 'prior_end_tp_idx' (not inclusive),
        'prior_mean', 'prior_sd', and 'prior_regressor_col'
    level_knot_dates : array like
        list of pre-specified dates for level knots
    level_knots : array like
        list of knot locations for level
        level_knot_dates and level_knots should be of the same length
    seasonal_knots_input : dict
         a dictionary for seasonality inputs with the following keys:
            '_seas_coef_knot_dates' : knot dates for seasonal regressors
            '_sea_coef_knot' : knot locations for sesonal regressors
            '_seasonality' : seasonality order
            '_seasonality_fs_order' : fourier series order for seasonality
    coefficients_knot_length : int
        the distance between every two knots for coefficients
    coefficients_knot_dates : array like
        a list of pre-specified knot dates for coefficients
    date_freq : str
        date frequency; if not supplied, pd.infer_freq will be used to imply the date frequency.
    min_residuals_sd : float
        a numeric value from 0 to 1 to indicate the lower bound of residual scale parameter; e.g.
        0.5 means residual scale will be sampled from [0, 0.5] in a scaled Beta(2, 2) dist.
    flat_multiplier : bool
        Default set as True. If False, we will adjust knot scale with a multiplier based on regressor volume
        around each knot; When True, set all multiplier as 1
    geometric_walk : bool
        Default set as False. If True we will sample positive regressor knot as geometric random walk
    kwargs
        To specify `estimator_type` or additional args for the specified `estimator_type`

    """
    _data_input_mapper = constants.DataInputMapper
    # stan or pyro model name (e.g. name of `*.stan` file in package)
    _model_name = 'ktrx'

    def __init__(self,
                 level_knot_scale=0.1,
                 regressor_col=None,
                 regressor_sign=None,
                 regressor_init_knot_loc=None,
                 regressor_init_knot_scale=None,
                 regressor_knot_scale=None,
                 span_coefficients=0.3,
                 rho_coefficients=0.15,
                 degree_of_freedom=30,
                 # time-based coefficient priors
                 coef_prior_list=None,
                 # knot customization
                 level_knot_dates=None,
                 level_knots=None,
                 seasonal_knots_input=None,
                 coefficients_knot_length=None,
                 coefficients_knot_dates=None,
                 date_freq=None,
                 mvn=0,
                 flat_multiplier=True,
                 geometric_walk=False,
                 min_residuals_sd=0.1,
                 **kwargs):
        super().__init__(**kwargs)  # create estimator in base class

        # normal distribution of known knot
        self.level_knot_scale = level_knot_scale

        self.level_knot_dates = level_knot_dates
        self._level_knot_dates = level_knot_dates

        self.level_knots = level_knots
        # self._level_knots = level_knots

        self._kernel_level = None
        self._num_knots_level = None
        self.knots_tp_level = None

        self._seasonal_knots_input = seasonal_knots_input
        self._seas_term = 0

        self.regressor_col = regressor_col
        self.regressor_sign = regressor_sign
        self.regressor_init_knot_loc = regressor_init_knot_loc
        self.regressor_init_knot_scale = regressor_init_knot_scale
        self.regressor_knot_scale = regressor_knot_scale

        self.coefficients_knot_length = coefficients_knot_length
        self.span_coefficients = span_coefficients
        self.rho_coefficients = rho_coefficients
        self.date_freq = date_freq

        self.degree_of_freedom = degree_of_freedom

        # multi var norm flag
        self.mvn = mvn
        # flat_multiplier flag
        self.flat_multiplier = flat_multiplier
        self.geometric_walk = geometric_walk
        self.min_residuals_sd = min_residuals_sd

        # set private var to arg value
        # if None set default in _set_default_args()
        self._regressor_sign = self.regressor_sign
        self._regressor_init_knot_loc = self.regressor_init_knot_loc
        self._regressor_init_knot_scale = self.regressor_init_knot_scale
        self._regressor_knot_scale = self.regressor_knot_scale

        self.coef_prior_list = coef_prior_list
        self._coef_prior_list = []
        self._coefficients_knot_dates = coefficients_knot_dates
        self._knots_idx_coef = None

        self._num_of_regressors = 0
        self._num_knots_coefficients = 0

        # positive regressors
        self._num_of_positive_regressors = 0
        self._positive_regressor_col = list()
        self._positive_regressor_init_knot_loc = list()
        self._positive_regressor_init_knot_scale = list()
        self._positive_regressor_knot_scale = list()
        # regular regressors
        self._num_of_regular_regressors = 0
        self._regular_regressor_col = list()
        self._regular_regressor_init_knot_loc = list()
        self._regular_regressor_init_knot_scale = list()
        self._regular_regressor_knot_scale = list()
        self._regressor_col = list()

        # init dynamic data attributes
        # the following are set by `_set_dynamic_attributes()` and generally set during fit()
        # from input df
        # response data
        self._is_valid_response = None
        self._which_valid_response = None
        self._num_of_valid_response = 0
        self._seasonality = None

        # regression data
        self._knots_tp_coefficients = None
        self._positive_regressor_matrix = None
        self._regular_regressor_matrix = None

    def _set_model_param_names(self):
        """Overriding base template functions. Model parameters to extract"""
        self._model_param_names += [param.value for param in constants.BaseSamplingParameters]
        if self._num_of_regressors > 0:
            self._model_param_names += [param.value for param in constants.RegressionSamplingParameters]

    def _set_default_args(self):
        """Set default attributes for None
        """
        if self.coef_prior_list is not None:
            self._coef_prior_list = deepcopy(self.coef_prior_list)
        if self.level_knots is None:
            self._level_knots = list()
        if self._seasonal_knots_input is not None:
            self._seasonality = self._seasonal_knots_input['_seasonality']
        else:
            self._seasonality = list()
        ##############################
        # if no regressors, end here #
        ##############################
        if self.regressor_col is None:
            # regardless of what args are set for these, if regressor_col is None
            # these should all be empty lists
            self._regressor_sign = list()
            self._regressor_init_knot_loc = list()
            self._regressor_init_knot_scale = list()
            self._regressor_knot_scale = list()

            return

        def _validate_params_len(params, valid_length):
            for p in params:
                if p is not None and len(p) != valid_length:
                    raise IllegalArgument('Wrong dimension length in Regression Param Input')

        # regressor defaults
        num_of_regressors = len(self.regressor_col)

        _validate_params_len([
            self.regressor_sign, self.regressor_init_knot_loc,
            self.regressor_init_knot_scale, self.regressor_knot_scale],
            num_of_regressors
        )

        if self.regressor_sign is None:
            self._regressor_sign = [DEFAULT_REGRESSOR_SIGN] * num_of_regressors

        if self.regressor_init_knot_loc is None:
            self._regressor_init_knot_loc = [DEFAULT_COEFFICIENTS_INIT_KNOT_LOC] * num_of_regressors

        if self.regressor_init_knot_scale is None:
            self._regressor_init_knot_scale = [DEFAULT_COEFFICIENTS_INIT_KNOT_SCALE] * num_of_regressors

        if self.regressor_knot_scale is None:
            self._regressor_knot_scale = [DEFAULT_COEFFICIENTS_KNOT_SCALE] * num_of_regressors

        self._num_of_regressors = num_of_regressors

    def _set_static_regression_attributes(self):
        # if no regressors, end here
        if self._num_of_regressors == 0:
            return

        for index, reg_sign in enumerate(self._regressor_sign):
            if reg_sign == '+':
                self._num_of_positive_regressors += 1
                self._positive_regressor_col.append(self.regressor_col[index])
                # used for 'pr_knot_loc' sampling in pyro
                self._positive_regressor_init_knot_loc.append(self._regressor_init_knot_loc[index])
                self._positive_regressor_init_knot_scale.append(self._regressor_init_knot_scale[index])
                # used for 'pr_knot' sampling in pyro
                self._positive_regressor_knot_scale.append(self._regressor_knot_scale[index])
            else:
                self._num_of_regular_regressors += 1
                self._regular_regressor_col.append(self.regressor_col[index])
                # used for 'rr_knot_loc' sampling in pyro
                self._regular_regressor_init_knot_loc.append(self._regressor_init_knot_loc[index])
                self._regular_regressor_init_knot_scale.append(self._regressor_init_knot_scale[index])
                # used for 'rr_knot' sampling in pyro
                self._regular_regressor_knot_scale.append(self._regressor_knot_scale[index])
        # regular first, then positive
        self._regressor_col = self._regular_regressor_col + self._positive_regressor_col
        # numpy conversion
        self._positive_regressor_init_knot_loc = np.array(self._positive_regressor_init_knot_loc)
        self._positive_regressor_init_knot_scale = np.array(self._positive_regressor_init_knot_scale)
        self._positive_regressor_knot_scale = np.array(self._positive_regressor_knot_scale)
        self._regular_regressor_init_knot_loc = np.array(self._regular_regressor_init_knot_loc)
        self._regular_regressor_init_knot_scale = np.array(self._regular_regressor_init_knot_scale)
        self._regular_regressor_knot_scale = np.array(self._regular_regressor_knot_scale)

    @staticmethod
    def _validate_coef_prior(coef_prior_list):
        for test_dict in coef_prior_list:
            if set(test_dict.keys()) != set([
                CoefPriorDictKeys.NAME.value,
                CoefPriorDictKeys.PRIOR_START_TP_IDX.value,
                CoefPriorDictKeys.PRIOR_END_TP_IDX.value,
                CoefPriorDictKeys.PRIOR_MEAN.value,
                CoefPriorDictKeys.PRIOR_SD.value,
                CoefPriorDictKeys.PRIOR_REGRESSOR_COL.value
            ]):
                raise IllegalArgument('wrong key name in inserted prior dict')
            len_insert_prior = list()
            for key, val in test_dict.items():
                if key in [
                    CoefPriorDictKeys.PRIOR_MEAN.value,
                    CoefPriorDictKeys.PRIOR_SD.value,
                    CoefPriorDictKeys.PRIOR_REGRESSOR_COL.value,
                ]:
                    len_insert_prior.append(len(val))
            if not all(len_insert == len_insert_prior[0] for len_insert in len_insert_prior):
                raise IllegalArgument('wrong dimension length in inserted prior dict')

    @staticmethod
    def _validate_level_knot_inputs(level_knot_dates, level_knots):
        if len(level_knots) != len(level_knot_dates):
            raise IllegalArgument('level_knots and level_knot_dates should have the same length')

    @staticmethod
    def _get_gap_between_dates(start_date, end_date, freq):
        diff = end_date - start_date
        gap = np.array(diff / np.timedelta64(1, freq))

        return gap

    @staticmethod
    def _set_knots_tp(knots_distance, cutoff):
        """provide a array like outcome of index based on the knots distance and cutoff point"""
        # start in the middle
        knots_idx_start = round(knots_distance / 2)
        knots_idx = np.arange(knots_idx_start, cutoff, knots_distance)

        return knots_idx

    def _set_coef_prior_idx(self):
        if self._coef_prior_list and len(self._regressor_col) > 0:
            for x in self._coef_prior_list:
                prior_regressor_col_idx = [
                    np.where(np.array(self._regressor_col) == col)[0][0]
                    for col in x['prior_regressor_col']
                ]
                x.update({'prior_regressor_col_idx': prior_regressor_col_idx})

    def _set_static_attributes(self):
        """model data input based on args at instantiation or computed from args at instantiation"""
        self._set_default_args()
        self._set_static_regression_attributes()

        self._validate_level_knot_inputs(self.level_knot_dates, self.level_knots)

        if self._coef_prior_list:
            self._validate_coef_prior(self._coef_prior_list)
            self._set_coef_prior_idx()

    def _set_valid_response_attributes(self):
        if self._seasonality:
            max_seasonality = np.round(np.max(self._seasonality)).astype(int)
            if self.num_of_observations < max_seasonality:
                raise ModelException(
                    "Number of observations {} is less than max seasonality {}".format(
                        self.num_of_observations, max_seasonality))
        # get some reasonable offset to regularize response to make default priors scale-insensitive
        if self._seasonality:
            max_seasonality = np.round(np.max(self._seasonality)).astype(int)
            self.response_offset = np.nanmean(self.response[:max_seasonality])
        else:
            self.response_offset = np.nanmean(self.response)

        self.is_valid_response = ~np.isnan(self.response)
        # [0] to convert tuple back to array
        self.which_valid_response = np.where(self.is_valid_response)[0]
        self.num_of_valid_response = len(self.which_valid_response)

    def _set_regressor_matrix(self, df):
        # validate regression columns
        if self.regressor_col is not None and \
                not set(self.regressor_col).issubset(df.columns):
            raise ModelException(
                "DataFrame does not contain specified regressor column(s)."
            )

        # init of regression matrix depends on length of response vector
        self._positive_regressor_matrix = np.zeros((self.num_of_observations, 0), dtype=np.double)
        self._regular_regressor_matrix = np.zeros((self.num_of_observations, 0), dtype=np.double)

        # update regression matrices
        if self._num_of_positive_regressors > 0:
            self._positive_regressor_matrix = df.filter(
                items=self._positive_regressor_col,).values

        if self._num_of_regular_regressors > 0:
            self._regular_regressor_matrix = df.filter(
                items=self._regular_regressor_col,).values

    def _set_coefficients_kernel_matrix(self, df):
        """Derive knots position and kernel matrix and other related meta data"""
        # Note that our tp starts by 1; to convert back to index of array, reduce it by 1
        tp = np.arange(1, self.num_of_observations + 1) / self.num_of_observations
        # this approach put knots in full range
        # TODO: consider deprecate _cutoff for now since we assume _cutoff always the same as num of obs?
        self._cutoff = self.num_of_observations
        self._kernel_coefficients = np.zeros((self.num_of_observations, 0), dtype=np.double)
        self._num_knots_coefficients = 0

        # kernel of coefficients calculations
        if self._num_of_regressors > 0:
            # if users didn't provide knot positions, evenly distribute it based on span_coefficients
            # or knot length provided by users
            if self._coefficients_knot_dates is None:
                # original code
                if self.coefficients_knot_length is not None:
                    knots_distance = self.coefficients_knot_length
                else:
                    number_of_knots = round(1 / self.span_coefficients)
                    knots_distance = math.ceil(self._cutoff / number_of_knots)
                # derive actual date arrays based on the time-point (tp) index
                knots_idx_coef = self._set_knots_tp(knots_distance, self._cutoff)
                self._knots_tp_coefficients = (1 + knots_idx_coef) / self.num_of_observations
                self._coefficients_knot_dates = df[self.date_col].values[knots_idx_coef]
                self._knots_idx_coef = knots_idx_coef
                # TODO: new idea
                # # ignore this case for now
                # # if self.coefficients_knot_length is not None:
                # #     knots_distance = self.coefficients_knot_length
                # # else:
                # number_of_knots = round(1 / self.span_coefficients)
                # # to work with index; has to be discrete
                # knots_distance = math.ceil(self._cutoff / number_of_knots)
                # # always has a knot at the starting point
                # # derive actual date arrays based on the time-point (tp) index
                # knots_idx_coef = np.arange(0, self._cutoff, knots_distance)
                # self._knots_tp_coefficients = (1 + knots_idx_coef) / self.num_of_observations
                # self._coefficients_knot_dates = df[self.date_col].values[knots_idx_coef]
                # self._knots_idx_coef = knots_idx_coef
            else:
                # FIXME: this only works up to daily series (not working on hourly series)
                # FIXME: didn't provide  self.knots_idx_coef in this case
                self._coefficients_knot_dates = pd.to_datetime([
                    x for x in self._coefficients_knot_dates if (x <= df[self.date_col].max()) \
                                                                and (x >= df[self.date_col].min())
                ])
                if self.date_freq is None:
                    self.date_freq = pd.infer_freq(df[self.date_col])[0]
                start_date = self.training_start
                self._knots_idx_coef = (
                    self._get_gap_between_dates(start_date, self._coefficients_knot_dates, self.date_freq)
                )

                self._knots_tp_coefficients = np.array(
                    (self._knots_idx_coef + 1) /
                    (self._get_gap_between_dates(start_date, self.training_end, self.date_freq) + 1)
                )
                self._knots_idx_coef = list(self._knots_idx_coef.astype(np.int32))

            kernel_coefficients = gauss_kernel(tp, self._knots_tp_coefficients, rho=self.rho_coefficients)

            self._num_knots_coefficients = len(self._knots_tp_coefficients)
            self._kernel_coefficients = kernel_coefficients

    def _set_knots_scale_matrix(self):
        if self._num_of_positive_regressors > 0:
            # calculate average local absolute volume for each segment
            local_val = np.ones((self._num_of_positive_regressors, self._num_knots_coefficients))
            if self.flat_multiplier:
                multiplier = np.ones(local_val.shape)
            else:
                multiplier = np.ones(local_val.shape)
                # store local value for the range on the left side since last knot
                for idx in range(len(self._knots_idx_coef)):
                    if idx < len(self._knots_idx_coef) - 1:
                        str_idx = self._knots_idx_coef[idx]
                        end_idx = self._knots_idx_coef[idx + 1]
                    else:
                        str_idx = self._knots_idx_coef[idx]
                        end_idx = self.num_of_observations

                    local_val[:, idx] = np.mean(np.fabs(self._positive_regressor_matrix[str_idx:end_idx]), axis=0)

                # adjust knot scale with the multiplier derive by the average value and shift by 0.001 to avoid zeros in
                # scale parameters
                global_mean = np.expand_dims(np.mean(np.fabs(self._positive_regressor_matrix), axis=0), -1)
                test_flag = local_val < 0.01 * global_mean

                multiplier[test_flag] = DEFAULT_LOWER_BOUND_SCALE_MULTIPLIER
                # replace entire row of nan (when 0.1 * global_mean is equal to global_min) with upper bound
                multiplier[np.isnan(multiplier).all(axis=-1)] = 1.0

            # also note that after the following step,
            # _positive_regressor_knot_scale is a 2D array unlike _regular_regressor_knot_scale
            # geometric drift i.e. 0.1 = 10% up-down in 1 s.d. prob.
            # after line below, self._positive_regressor_knot_scale has shape num_of_pr x num_of_knot
            self._positive_regressor_knot_scale = (
                    multiplier * np.expand_dims(self._positive_regressor_knot_scale, -1)
            )
            # keep a lower bound of scale parameters
            self._positive_regressor_knot_scale[self._positive_regressor_knot_scale < 1e-4] = 1e-4
            # TODO: we change the type here, maybe we should change it earlier?
            self._positive_regressor_init_knot_scale = np.array(self._positive_regressor_init_knot_scale)
            self._positive_regressor_init_knot_scale[self._positive_regressor_init_knot_scale < 1e-4] = 1e-4

        if self._num_of_regular_regressors > 0:
            # do the same for regular regressor
            # calculate average local absolute volume for each segment
            local_val = np.ones((self._num_of_regular_regressors, self._num_knots_coefficients))
            if self.flat_multiplier:
                multiplier = np.ones(local_val.shape)
            else:
                multiplier = np.ones(local_val.shape)
            # store local value for the range on the left side since last knot
            for idx in range(len(self._knots_idx_coef)):
                if idx < len(self._knots_idx_coef) - 1:
                    str_idx = self._knots_idx_coef[idx]
                    end_idx = self._knots_idx_coef[idx + 1]
                else:
                    str_idx = self._knots_idx_coef[idx]
                    end_idx = self.num_of_observations

                local_val[:, idx] = np.mean(np.fabs(self._regular_regressor_matrix[str_idx:end_idx]), axis=0)

            # adjust knot scale with the multiplier derive by the average value and shift by 0.001 to avoid zeros in
            # scale parameters
            global_mean = np.expand_dims(np.mean(np.fabs(self._regular_regressor_matrix), axis=0), -1)
            test_flag = local_val < 0.01 * global_mean
            multiplier[test_flag] = DEFAULT_LOWER_BOUND_SCALE_MULTIPLIER
            # replace entire row of nan (when 0.1 * global_mean is equal to global_min) with upper bound
            multiplier[np.isnan(multiplier).all(axis=-1)] = 1.0

            # also note that after the following step,
            # _regular_regressor_knot_scale is a 2D array unlike _regular_regressor_knot_scale
            # geometric drift i.e. 0.1 = 10% up-down in 1 s.d. prob.
            # self._regular_regressor_knot_scale has shape num_of_pr x num_of_knot
            self._regular_regressor_knot_scale = (
                    multiplier * np.expand_dims(self._regular_regressor_knot_scale, -1)
            )
            # keep a lower bound of scale parameters
            self._regular_regressor_knot_scale[self._regular_regressor_knot_scale < 1e-4] = 1e-4
            # TODO: we change the type here, maybe we should change it earlier?
            self._regular_regressor_init_knot_scale = np.array(self._regular_regressor_init_knot_scale)
            self._regular_regressor_init_knot_scale[self._regular_regressor_init_knot_scale < 1e-4] = 1e-4

    def _generate_tp(self, prediction_date_array):
        """Used in _generate_coefs"""
        prediction_start = prediction_date_array[0]
        output_len = len(prediction_date_array)
        if prediction_start > self.training_end:
            start = self.num_of_observations
        else:
            start = pd.Index(self.date_array).get_loc(prediction_start)

        new_tp = np.arange(start + 1, start + output_len + 1) / self.num_of_observations
        return new_tp

    def _generate_insample_tp(self, date_array):
        """Used in _generate_coefs"""
        idx = np.nonzero(np.in1d(self.date_array, date_array))[0]
        tp = (idx + 1) / self.num_of_observations
        return tp

    def _generate_coefs(self, prediction_date_array, coef_knot_dates, coef_knot):
        """Used in _generate_seas"""
        new_tp = self._generate_tp(prediction_date_array)
        knots_tp_coef = self._generate_insample_tp(coef_knot_dates)
        kernel_coef = sandwich_kernel(new_tp, knots_tp_coef)
        coefs = np.squeeze(np.matmul(coef_knot, kernel_coef.transpose(1, 0)), axis=0).transpose(1, 0)
        return coefs

    def _generate_seas(self, df, coef_knot_dates, coef_knot, seasonality, seasonality_fs_order):
        """To calculate the seasonality term based on the _seasonal_knots_input.

        :param df: input df
        :param coef_knot_dates: dates for coef knots
        :param coef_knot: knot values for coef
        :param seasonality: seasonality input
        :param seasonality_fs_order: seasonality_fs_order input
        :return:
        """
        prediction_date_array = df[self.date_col].values
        prediction_start = prediction_date_array[0]

        df = df.copy()
        if prediction_start > self.training_end:
            forecast_dates = set(prediction_date_array)
            n_forecast_steps = len(forecast_dates)
            # time index for prediction start
            start = self.num_of_observations
        else:
            # compute how many steps to forecast
            forecast_dates = set(prediction_date_array) - set(self.date_array)
            # check if prediction df is a subset of training df
            # e.g. "negative" forecast steps
            n_forecast_steps = len(forecast_dates) or \
                               - (len(set(self.date_array) - set(prediction_date_array)))
            # time index for prediction start
            start = pd.Index(self.date_array).get_loc(prediction_start)

        fs_cols = []
        for idx, s in enumerate(seasonality):
            order = seasonality_fs_order[idx]
            df, fs_cols_temp = make_fourier_series_df(df, s, order=order, prefix='seas{}_'.format(s), shift=start)
            fs_cols += fs_cols_temp

        sea_regressor_matrix = df.filter(items=fs_cols).values
        sea_coefs = self._generate_coefs(prediction_date_array, coef_knot_dates, coef_knot)
        seas = np.sum(sea_coefs * sea_regressor_matrix, axis=-1)

        return seas

    def _set_levs_and_seas(self, df):
        tp = np.arange(1, self.num_of_observations + 1) / self.num_of_observations
        # trim level knots dates when they are beyond training dates
        lev_knot_dates = list()
        lev_knots = list()
        # TODO: any faster way instead of a simple loop?
        for i, x in enumerate(self.level_knot_dates):
            if (x <= df[self.date_col].max()) and (x >= df[self.date_col].min()):
                lev_knot_dates.append(x)
                lev_knots.append(self.level_knots[i])
        self._level_knot_dates = pd.to_datetime(lev_knot_dates)
        self._level_knots = np.array(lev_knots)
        infer_freq = pd.infer_freq(df[self.date_col])[0]
        start_date = self.training_start

        if len(self.level_knots) > 0 and len(self.level_knot_dates) > 0:
            self.knots_tp_level = np.array(
                (self._get_gap_between_dates(start_date, self._level_knot_dates, infer_freq) + 1) /
                (self._get_gap_between_dates(start_date, self.training_end, infer_freq) + 1)
            )
        else:
            raise ModelException("User need to supply a list of level knots.")

        kernel_level = sandwich_kernel(tp, self.knots_tp_level)
        self._kernel_level = kernel_level
        self._num_knots_level = len(self._level_knot_dates)

        if self._seasonal_knots_input is not None:
            self._seas_term = self._generate_seas(
                df,
                self._seasonal_knots_input['_seas_coef_knot_dates'],
                self._seasonal_knots_input['_sea_coef_knot'],
                # self._seasonal_knots_input['_sea_rho'],
                self._seasonal_knots_input['_seasonality'],
                self._seasonal_knots_input['_seasonality_fs_order'])

    def _filter_coef_prior(self, df):
        if self._coef_prior_list and len(self._regressor_col) > 0:
            # iterate over a copy due to the removal operation
            for test_dict in self._coef_prior_list[:]:
                prior_regressor_col = test_dict['prior_regressor_col']
                m = test_dict['prior_mean']
                sd = test_dict['prior_sd']
                end_tp_idx = min(test_dict['prior_end_tp_idx'], df.shape[0])
                start_tp_idx = min(test_dict['prior_start_tp_idx'], df.shape[0])
                if start_tp_idx < end_tp_idx:
                    expected_shape = (end_tp_idx - start_tp_idx, len(prior_regressor_col))
                    test_dict.update({'prior_end_tp_idx': end_tp_idx})
                    test_dict.update({'prior_start_tp_idx': start_tp_idx})
                    # mean/sd expanding
                    test_dict.update({'prior_mean': np.full(expected_shape, m)})
                    test_dict.update({'prior_sd': np.full(expected_shape, sd)})
                else:
                    # removing invalid prior
                    self._coef_prior_list.remove(test_dict)

    def _set_dynamic_attributes(self, df):
        """Overriding: func: `~orbit.models.BaseETS._set_dynamic_attributes"""
        self._set_valid_response_attributes()
        self._set_regressor_matrix(df)
        self._set_coefficients_kernel_matrix(df)
        self._set_knots_scale_matrix()
        self._set_levs_and_seas(df)
        self._filter_coef_prior(df)

    @staticmethod
    def _concat_regression_coefs(pr_beta=None, rr_beta=None):
        """Concatenates regression posterior matrix

        In the case that `pr_beta` or `rr_beta` is a 1d tensor, transform to 2d tensor and
        concatenate.

        Args
        ----
        pr_beta : array like
            postive-value constrainted regression betas
        rr_beta : array like
            regular regression betas

        Returns
        -------
        array like
            concatenated 2d array of shape (1, len(rr_beta) + len(pr_beta))

        """
        regressor_beta = None
        if pr_beta is not None and rr_beta is not None:
            pr_beta = pr_beta if len(pr_beta.shape) == 2 else pr_beta.reshape(1, -1)
            rr_beta = rr_beta if len(rr_beta.shape) == 2 else rr_beta.reshape(1, -1)
            regressor_beta = torch.cat((rr_beta, pr_beta), dim=1)
        elif pr_beta is not None:
            regressor_beta = pr_beta
        elif rr_beta is not None:
            regressor_beta = rr_beta

        return regressor_beta

    def _predict(self, posterior_estimates, df, include_error=False, decompose=False, store_prediction_array=False,
                 coefficient_method="smooth",
                 **kwargs):
        """Vectorized version of prediction math

        Args
        ----
        coefficient_method: str
            either "smooth" or "empirical". when "empirical" is used, curves are sampled/aggregated directly
            from beta posteriors; when "smooth" is used, first extract sampled/aggregated posteriors of knots
            then beta.
            this mainly impacts the aggregated estimation method; full bayesian should not be impacted
        """

        # remove reference from original input
        df = df.copy()
        prediction_df_meta = self.get_prediction_df_meta(df)

        ################################################################
        # Model Attributes
        ################################################################

        model = deepcopy(posterior_estimates)
        arbitrary_posterior_value = list(model.values())[0]
        num_sample = arbitrary_posterior_value.shape[0]

        ################################################################
        # Prediction Attributes
        ################################################################
        output_len = prediction_df_meta['df_length']
        prediction_start = prediction_df_meta['prediction_start']

        # Here assume dates are ordered and consecutive
        # if prediction_df_meta['prediction_start'] > self.training_end,
        # assume prediction starts right after train end
        if prediction_start > self.training_end:
            # time index for prediction start
            start = self.num_of_observations
        else:
            start = pd.Index(self.date_array).get_loc(prediction_start)

        new_tp = np.arange(start + 1, start + output_len + 1) / self.num_of_observations
        if include_error:
            # in-sample knots
            lev_knot_in = model.get(constants.BaseSamplingParameters.LEVEL_KNOT.value)
            # TODO: hacky way; let's just assume last two knot distance is knots distance for all knots
            lev_knot_width = self.knots_tp_level[-1] - self.knots_tp_level[-2]
            # check whether we need to put new knots for simulation
            if new_tp[-1] >= self.knots_tp_level[-1] + lev_knot_width:
                # derive knots tp
                knots_tp_level_out = np.arange(self.knots_tp_level[-1] + lev_knot_width, new_tp[-1], lev_knot_width)
                new_knots_tp_level = np.concatenate([self.knots_tp_level, knots_tp_level_out])
                lev_knot_out = np.random.laplace(0, self.level_knot_scale,
                                                 size=(lev_knot_in.shape[0], len(knots_tp_level_out)))
                lev_knot_out = np.cumsum(np.concatenate([lev_knot_in[:, -1].reshape(-1, 1), lev_knot_out],
                                                        axis=1), axis=1)[:, 1:]
                lev_knot = np.concatenate([lev_knot_in, lev_knot_out], axis=1)
            else:
                new_knots_tp_level = self.knots_tp_level
                lev_knot = lev_knot_in
            kernel_level = sandwich_kernel(new_tp, new_knots_tp_level)
        else:
            lev_knot = model.get(constants.BaseSamplingParameters.LEVEL_KNOT.value)
            kernel_level = sandwich_kernel(new_tp, self.knots_tp_level)
        obs_scale = model.get(constants.BaseSamplingParameters.OBS_SCALE.value)
        obs_scale = obs_scale.reshape(-1, 1)

        if self._seasonal_knots_input is not None:
            seas = self._generate_seas(df,
                                       self._seasonal_knots_input['_seas_coef_knot_dates'],
                                       self._seasonal_knots_input['_sea_coef_knot'],
                                       # self._seasonal_knots_input['_sea_rho'],
                                       self._seasonal_knots_input['_seasonality'],
                                       self._seasonal_knots_input['_seasonality_fs_order'])
            # seas is 1-d array, add the batch size back
            seas = np.expand_dims(seas, 0)
        else:
            # follow component shapes
            seas = np.zeros((1, output_len))

        trend = np.matmul(lev_knot, kernel_level.transpose((1, 0)))
        regression = np.zeros(trend.shape)
        if self._num_of_regressors > 0:
            regressor_matrix = df.filter(items=self._regressor_col,).values
            regressor_betas = self._get_regression_coefs_matrix(
                model, coefficient_method, prediction_df_meta['date_array']
            )
            regression = np.sum(regressor_betas * regressor_matrix, axis=-1)

        if include_error:
            epsilon = nct.rvs(self.degree_of_freedom, nc=0, loc=0,
                              scale=obs_scale, size=(num_sample, len(new_tp)))
            pred_array = trend + seas + regression + epsilon
        else:
            pred_array = trend + seas + regression

        # if decompose output dictionary of components
        if decompose:
            decomp_dict = {
                'prediction': pred_array,
                'trend': trend,
                'seasonality_input': seas,
                'regression': regression
            }
        else:
            decomp_dict = {'prediction': pred_array}

        if store_prediction_array:
            self.pred_array = pred_array
        else:
            self.pred_array = None

        return decomp_dict

    def _get_regression_coefs_matrix(self, model, coefficient_method='smooth', date_array=None):
        """internal function to provide coefficient matrix given a date array

        Args
        ----
        model: dict
            posterior samples
        date_array: array like
            array of date stamp
        coefficient_method: str
            either "empirical" or "smooth"; when "empirical" is used; curve are sampled/aggregated directly from
            coefficients posteriors whereas when "smooth" is used we first extract sampled/aggregated posteriors of knot
            and extract coefficients this mainly impact the aggregated estimation method; full bayesian should not be
            impacted
        """
        if self._num_of_regular_regressors + self._num_of_positive_regressors == 0:
            return None

        if date_array is None:
            if coefficient_method == 'smooth':
                # if date_array not specified, dynamic coefficients in the training perior will be retrieved
                coef_knots = model.get(constants.RegressionSamplingParameters.COEFFICIENTS_KNOT.value)
                regressor_betas = np.matmul(coef_knots, self._kernel_coefficients.transpose((1, 0)))
                # back to batch x time step x regressor columns shape
                regressor_betas = regressor_betas.transpose((0, 2, 1))
            elif coefficient_method == 'empirical':
                regressor_betas = model.get(constants.RegressionSamplingParameters.COEFFICIENTS.value)
            else:
                raise IllegalArgument('Wrong coefficient_method:{}'.format(coefficient_method))
        else:
            date_array = pd.to_datetime(date_array).values
            output_len = len(date_array)
            train_len = self.num_of_observations
            # some validation of date array
            if not is_ordered_datetime(date_array):
                raise IllegalArgument('Datetime index must be ordered and not repeat')
            prediction_start = date_array[0]

            if prediction_start < self.training_start:
                raise PredictionException('Prediction start must be after training start.')

            # If we cannot find a match of prediction range, assume prediction starts right after train end
            if prediction_start > self.training_end:
                # time index for prediction start
                start = train_len
                coef_repeats = [0] * (start - 1) + [output_len]
            else:
                # time index for prediction start
                start = pd.Index(self.date_array).get_loc(prediction_start)
                if output_len <= train_len - start:
                    coef_repeats = [0] * start + [1] * output_len + [0] * (train_len - start - output_len)
                else:
                    coef_repeats = [0] * start + [1] * (train_len - start - 1) + [output_len - train_len + start + 1]
            new_tp = np.arange(start + 1, start + output_len + 1) / self.num_of_observations

            if coefficient_method == 'smooth':
                kernel_coefficients = gauss_kernel(new_tp, self._knots_tp_coefficients, rho=self.rho_coefficients)
                # kernel_coefficients = parabolic_kernel(new_tp, self._knots_tp_coefficients)
                coef_knots = model.get(constants.RegressionSamplingParameters.COEFFICIENTS_KNOT.value)
                regressor_betas = np.matmul(coef_knots, kernel_coefficients.transpose((1, 0)))
                regressor_betas = regressor_betas.transpose((0, 2, 1))
            elif coefficient_method == 'empirical':
                regressor_betas = model.get(constants.RegressionSamplingParameters.COEFFICIENTS.value)
                regressor_betas = np.repeat(regressor_betas, repeats=coef_repeats, axis=1)
            else:
                raise IllegalArgument('Wrong coefficient_method:{}'.format(coefficient_method))

        return regressor_betas

    def _get_regression_coefs(self, aggregate_method, coefficient_method='smooth', date_array=None,
                              include_ci=False, lower=0.05, upper=0.95):
        """Return DataFrame regression coefficients
        """
        posteriors = self._aggregate_posteriors.get(aggregate_method)
        coefs = np.squeeze(self._get_regression_coefs_matrix(posteriors,
                                                             coefficient_method=coefficient_method,
                                                             date_array=date_array))
        if len(coefs.shape) == 1:
            coefs = coefs.reshape((1, -1))
        reg_df = pd.DataFrame(data=coefs, columns=self._regressor_col)
        if date_array is not None:
            reg_df[self.date_col] = date_array
        else:
            reg_df[self.date_col] = self.date_array

        # re-arrange columns
        reg_df = reg_df[[self.date_col] + self._regressor_col]
        if include_ci:
            posteriors = self._posterior_samples
            coefs = self._get_regression_coefs_matrix(posteriors, coefficient_method=coefficient_method)

            coefficients_lower = np.quantile(coefs, lower, axis=0)
            coefficients_upper = np.quantile(coefs, upper, axis=0)

            reg_df_lower = reg_df.copy()
            reg_df_upper = reg_df.copy()
            for idx, col in enumerate(self._regressor_col):
                reg_df_lower[col] = coefficients_lower[:, idx]
                reg_df_upper[col] = coefficients_upper[:, idx]
            return reg_df, reg_df_lower, reg_df_upper

        return reg_df

    def _get_regression_coef_knots(self, aggregate_method):
        """Return DataFrame regression coefficient knots
        """
        # init dataframe
        knots_df = pd.DataFrame()
        # end if no regressors
        if self._num_of_regular_regressors + self._num_of_positive_regressors == 0:
            return knots_df

        knots_df[self.date_col] = self._coefficients_knot_dates
        # TODO: make the label as a constant
        knots_df['step'] = self._knots_idx_coef
        coef_knots = self._aggregate_posteriors \
            .get(aggregate_method) \
            .get(constants.RegressionSamplingParameters.COEFFICIENTS_KNOT.value)

        for idx, col in enumerate(self._regressor_col):
            knots_df[col] = np.transpose(coef_knots[:, idx])

        return knots_df

    @staticmethod
    def _plot_regression_coefs(coef_df, knot_df=None, coef_df_lower=None, coef_df_upper=None,
                               ncol=2, figsize=None, ylim=None, markersize=200):
        """Plot regression coefficients
        """
        # assume your first column is the date; this way can use a static method
        regressor_col = coef_df.columns.tolist()[1:]
        nrow = math.ceil(len(regressor_col) / ncol)
        fig, axes = plt.subplots(nrow, ncol, figsize=figsize, squeeze=False)

        for idx, col in enumerate(regressor_col):
            row_idx = idx // ncol
            col_idx = idx % ncol
            coef = coef_df[col]
            axes[row_idx, col_idx].plot(coef, alpha=.8, label='coefficients')
            if coef_df_lower is not None and coef_df_upper is not None:
                coef_lower = coef_df_lower[col]
                coef_upper = coef_df_upper[col]
                axes[row_idx, col_idx].fill_between(np.arange(0, coef_df.shape[0]), coef_lower, coef_upper, alpha=.3)
            if knot_df is not None:
                step = knot_df['step']
                knots = knot_df[col].values
                axes[row_idx, col_idx].scatter(x=step, y=knots, marker='^', s=markersize, color='green', alpha=0.5)
            if ylim is not None:
                axes[row_idx, col_idx].set_ylim(ylim)
            axes[row_idx, col_idx].set_title('{}'.format(col))
            axes[row_idx, col_idx].ticklabel_format(useOffset=False)

        plt.tight_layout()
        return axes


class KTRXFull(FullBayesianTemplate, BaseKTRX):
    """Concrete KTRX model for full Bayesian prediction"""
    _supported_estimator_types = [PyroEstimatorVI]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_regression_coefs(self, aggregate_method='median',
                             coefficient_method='smooth',
                             date_array=None,
                             include_ci=False, lower=0.05, upper=0.95):
        self._set_aggregate_posteriors()
        return self._get_regression_coefs(aggregate_method=aggregate_method,
                                          coefficient_method=coefficient_method,
                                          date_array=date_array,
                                          include_ci=include_ci,
                                          lower=lower, upper=upper)

    def get_regression_coef_knots(self, aggregate_method='median'):
        self._set_aggregate_posteriors()
        return self._get_regression_coef_knots(aggregate_method=aggregate_method)

    def plot_regression_coefs(self, with_knot=False, aggregate_method='median', include_ci=False, **kwargs):
        if with_knot:
            knot_df = self.get_regression_coef_knots(aggregate_method=aggregate_method)
        else:
            knot_df = None
        if include_ci:
            coef_df, coef_df_lower, coef_df_upper = self.get_regression_coefs(
                aggregate_method=aggregate_method,
                include_ci=True
            )
        else:
            coef_df = self.get_regression_coefs(aggregate_method=aggregate_method,
                                                include_ci=False)
            coef_df_lower = None
            coef_df_upper = None
        return self._plot_regression_coefs(coef_df=coef_df,
                                           knot_df=knot_df,
                                           coef_df_lower=coef_df_lower,
                                           coef_df_upper=coef_df_upper,
                                           **kwargs)


class KTRXAggregated(AggregatedPosteriorTemplate, BaseKTRX):
    """Concrete KTRX model for aggregated Bayesian prediction"""
    _supported_estimator_types = [PyroEstimatorVI]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_regression_coefs(self, coefficient_method='smooth', date_array=None):
        return self._get_regression_coefs(aggregate_method=self.aggregate_method,
                                          coefficient_method=coefficient_method,
                                          date_array=date_array,
                                          include_ci=False)

    def get_regression_coef_knots(self):
        return self._get_regression_coef_knots(aggregate_method=self.aggregate_method)

    def plot_regression_coefs(self, with_knot=False, coefficient_method='smooth', **kwargs):
        coef_df = self._get_regression_coefs(aggregate_method=self.aggregate_method,
                                             coefficient_method=coefficient_method,
                                             include_ci=False)
        if with_knot:
            knot_df = self.get_regression_coef_knots()
        else:
            knot_df = None
        return self._plot_regression_coefs(coef_df=coef_df, knot_df=knot_df, **kwargs)



from enum import Enum


class DataInputMapper(Enum):
    """
    mapping from object input to pyro input
    """
    # All of the following have default defined in DEFAULT_SLGT_FIT_ATTRIBUTES
    # ----------  Data Input ---------- #
    # observation related
    NUM_OF_OBSERVATIONS = 'N_OBS'
    RESPONSE = 'RESPONSE'
    NUM_OF_VALID_RESPONSE = 'N_VALID_RES'
    WHICH_VALID_RESPONSE = 'WHICH_VALID_RES'
    RESPONSE_SD = 'SDY'
    RESPONSE_OFFSET = 'MEAN_Y'
    _DEGREE_OF_FREEDOM = 'DOF'
    # ----------  Level  ---------- #
    NUM_KNOTS_LEVEL = 'N_KNOTS_LEV'
    LEVEL_KNOT_SCALE = 'LEV_KNOT_SCALE'
    KERNEL_LEVEL = 'K_LEV'
    # ----------  Regression  ---------- #
    NUM_KNOTS_COEFFICIENTS  = 'N_KNOTS_COEF'
    KERNEL_COEFFICIENTS = 'K_COEF'
    NUM_OF_REGRESSORS = 'P'
    REGRESSOR_MATRIX = 'REGRESSORS'
    COEFFICIENTS_INITIAL_KNOT_SCALE = 'COEF_INIT_KNOT_SCALE'
    COEFFICIENTS_KNOT_SCALE = 'COEF_KNOT_SCALE'


class BaseSamplingParameters(Enum):
    """
    The output sampling parameters related with base model.
    """
    LEVEL_KNOT = 'lev_knot'
    LEVEL = 'lev'
    YHAT = 'yhat'
    OBS_SCALE = 'obs_scale'


class RegressionSamplingParameters(Enum):
    """
    The output sampling parameters related with regression component.
    """
    COEFFICIENTS_KNOT = 'coef_knot'
    COEFFICIENTS = 'coef'



from abc import abstractmethod
import numpy as np
from ..utils.docstring_style import merge_numpy_docs_dedup
import custom_inherit as ci
ci.store["numpy_with_merge_dedup"] = merge_numpy_docs_dedup
ci.add_style("numpy_with_merge_dedup", merge_numpy_docs_dedup)


class BaseEstimator(object, metaclass=ci.DocInheritMeta(style="numpy_with_merge_dedup")):
    """Base Estimator class for both Stan and Pyro Estimator

    Parameters
    ----------
    seed : int
        seed number for initial random values
    verbose : bool
        If True (default), output all diagnostics messages from estimators

    """
    def __init__(self, seed=8888, verbose=False):
        self.seed = seed
        self.verbose = verbose

        # set random state
        np.random.seed(self.seed)

    @abstractmethod
    def fit(self, model_name, model_param_names, data_input, fitter=None, init_values=None):
        """

        Parameters
        ----------
        model_name : str
            name of model - used in mapping the right sampling file (stan/pyro/...)
        model_param_names : list
            list of strings of model parameters names to extract
        data_input : dict
            key-value pairs of data input as required by definition in samplers (stan/pyro/...)
        fitter :
            model object used for fitting; this will be used instead of model_name if supplied to search for
            model object
        init_values : float or np.array
            initial sampler value. If None, 'random' is used

        Returns
        -------
        posteriors : dict
            key value pairs where key is the model parameter name and value is `num_sample` x posterior values
        training_metrics : dict
            metrics and meta data related to the training process
        """
        raise NotImplementedError('Concrete fit() method must be implemented')



import pytest
import numpy as np
import pandas as pd

from orbit.estimators.pyro_estimator import PyroEstimator, PyroEstimatorVI, PyroEstimatorMAP
from orbit.estimators.stan_estimator import StanEstimator, StanEstimatorMCMC, StanEstimatorVI, StanEstimatorMAP
from orbit.models.ktrlite import KTRLiteMAP
from orbit.models.ktrx import BaseKTRX, KTRXFull, KTRXAggregated
from orbit.diagnostics.metrics import smape


@pytest.mark.parametrize("regressor_col", [None, ['a', 'b', 'c']])
def test_ktrx_pyro(make_daily_data, regressor_col):
    train_df, test_df, coef = make_daily_data

    ktrlite = KTRLiteMAP(
        response_col='response',
        date_col='date',
        seasonality=[7, 365.25],
        seasonality_fs_order=[2, 5],
        estimator_type=StanEstimatorMAP
    )
    ktrlite.fit(train_df)

    level_knot_dates = ktrlite._level_knot_dates
    level_knots = ktrlite._aggregate_posteriors['map']['lev_knot'][0]
    seasonal_knots_input = {
        '_seas_coef_knot_dates': ktrlite._coef_knot_dates,
        '_sea_coef_knot': ktrlite._aggregate_posteriors['map']['coef_knot'],
        '_seasonality': ktrlite._seasonality,
        '_seasonality_fs_order': ktrlite._seasonality_fs_order,
    }
    ktrx = KTRXAggregated(
        response_col='response',
        date_col='date',
        regressor_col=regressor_col,
        level_knot_dates=level_knot_dates,
        level_knots=level_knots,
        seasonal_knots_input=seasonal_knots_input,
        estimator_type=PyroEstimatorVI,
    )

    ktrx.fit(train_df)
    predict_df = ktrx.predict(test_df)

    expected_columns = ['date', 'prediction_5', 'prediction', 'prediction_95']
    expected_shape = (364, len(expected_columns))

    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns
    assert smape(test_df['response'].values, predict_df['prediction'].values) <= 0.5


@pytest.mark.parametrize("coefficients_knot_dates", [pd.date_range(start='2016-03-01', end='2019-01-01', freq='3M')])
def test_ktrx_coef_knot_dates(make_daily_data, coefficients_knot_dates):
    train_df, test_df, coef = make_daily_data

    ktrlite = KTRLiteMAP(
        response_col='response',
        date_col='date',
        seasonality=[7, 365.25],
        seasonality_fs_order=[2, 5],
        estimator_type=StanEstimatorMAP
    )
    ktrlite.fit(train_df)

    level_knot_dates = ktrlite._level_knot_dates
    level_knots = ktrlite._aggregate_posteriors['map']['lev_knot'][0]
    seasonal_knots_input = {
        '_seas_coef_knot_dates': ktrlite._coef_knot_dates,
        '_sea_coef_knot': ktrlite._aggregate_posteriors['map']['coef_knot'],
        '_seasonality': ktrlite._seasonality,
        '_seasonality_fs_order': ktrlite._seasonality_fs_order,
    }
    ktrx = KTRXAggregated(
        response_col='response',
        date_col='date',
        regressor_col=['a', 'b', 'c'],
        coefficients_knot_dates=coefficients_knot_dates,
        level_knot_dates=level_knot_dates,
        level_knots=level_knots,
        seasonal_knots_input=seasonal_knots_input,
        estimator_type=PyroEstimatorVI,
    )

    ktrx.fit(train_df)
    predict_df = ktrx.predict(test_df)

    expected_columns = ['date', 'prediction_5', 'prediction', 'prediction_95']
    expected_shape = (364, len(expected_columns))
    expected_num_parameters = 7

    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns
    assert len(ktrx._posterior_samples) == expected_num_parameters
    assert smape(test_df['response'].values, predict_df['prediction'].values) <= 0.5


@pytest.mark.parametrize("coefficients_knot_length", [90, 120])
def test_ktrx_coef_knot_distance(make_daily_data, coefficients_knot_length):
    train_df, test_df, coef = make_daily_data

    ktrlite = KTRLiteMAP(
        response_col='response',
        date_col='date',
        seasonality=[7, 365.25],
        seasonality_fs_order=[2, 5],
        estimator_type=StanEstimatorMAP
    )
    ktrlite.fit(train_df)

    level_knot_dates = ktrlite._level_knot_dates
    level_knots = ktrlite._aggregate_posteriors['map']['lev_knot'][0]
    seasonal_knots_input = {
        '_seas_coef_knot_dates': ktrlite._coef_knot_dates,
        '_sea_coef_knot': ktrlite._aggregate_posteriors['map']['coef_knot'],
        '_seasonality': ktrlite._seasonality,
        '_seasonality_fs_order': ktrlite._seasonality_fs_order,
    }
    ktrx = KTRXAggregated(
        response_col='response',
        date_col='date',
        regressor_col=['a', 'b', 'c'],
        coefficients_knot_length=coefficients_knot_length,
        level_knot_dates=level_knot_dates,
        level_knots=level_knots,
        seasonal_knots_input=seasonal_knots_input,
        estimator_type=PyroEstimatorVI,
    )

    ktrx.fit(train_df)
    predict_df = ktrx.predict(test_df)

    expected_columns = ['date', 'prediction_5', 'prediction', 'prediction_95']
    expected_shape = (364, len(expected_columns))
    expected_num_parameters = 7

    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns
    assert len(ktrx._posterior_samples) == expected_num_parameters
    assert smape(test_df['response'].values, predict_df['prediction'].values) <= 0.5


@pytest.mark.parametrize(
    "regressor_signs",
    [
        ["+", "+", "+"],
        ["=", "=", "="],
        ["+", "=", "+"]
    ],
    ids=['positive_only', 'regular_only', 'mixed_signs']
)
def test_ktrx_regressor_sign(make_daily_data, regressor_signs):
    train_df, test_df, coef = make_daily_data

    ktrlite = KTRLiteMAP(
        response_col='response',
        date_col='date',
        seasonality=[7, 365.25],
        seasonality_fs_order=[2, 5],
        estimator_type=StanEstimatorMAP
    )
    ktrlite.fit(train_df)

    level_knot_dates = ktrlite._level_knot_dates
    level_knots = ktrlite._aggregate_posteriors['map']['lev_knot'][0]
    seasonal_knots_input = {
        '_seas_coef_knot_dates': ktrlite._coef_knot_dates,
        '_sea_coef_knot': ktrlite._aggregate_posteriors['map']['coef_knot'],
        '_seasonality': ktrlite._seasonality,
        '_seasonality_fs_order': ktrlite._seasonality_fs_order,
    }
    ktrx = KTRXAggregated(
        response_col='response',
        date_col='date',
        regressor_col=['a', 'b', 'c'],
        regressor_sign=regressor_signs,
        level_knot_dates=level_knot_dates,
        level_knots=level_knots,
        seasonal_knots_input=seasonal_knots_input,
        estimator_type=PyroEstimatorVI,
    )

    ktrx.fit(train_df)
    predict_df = ktrx.predict(test_df)

    expected_columns = ['date', 'prediction_5', 'prediction', 'prediction_95']
    expected_shape = (364, len(expected_columns))
    expected_num_parameters = 7

    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns
    assert len(ktrx._posterior_samples) == expected_num_parameters
    assert smape(test_df['response'].values, predict_df['prediction'].values) <= 0.5


@pytest.mark.parametrize(
    "coef_prior_list",
    [
        [
            {
                'name': 'test1',
                'prior_start_tp_idx': 100,
                'prior_end_tp_idx': 120,
                'prior_mean': [.25, 0.35],
                'prior_sd': [.1, .2],
                'prior_regressor_col': ['a', 'b']

            }
        ]
    ]
)
def test_ktrx_prior_ingestion(make_daily_data, coef_prior_list):
    train_df, test_df, coef = make_daily_data

    ktrlite = KTRLiteMAP(
        response_col='response',
        date_col='date',
        seasonality=[7, 365.25],
        seasonality_fs_order=[2, 5],
        estimator_type=StanEstimatorMAP
    )
    ktrlite.fit(train_df)

    level_knot_dates = ktrlite._level_knot_dates
    level_knots = ktrlite._aggregate_posteriors['map']['lev_knot'][0]
    seasonal_knots_input = {
        '_seas_coef_knot_dates': ktrlite._coef_knot_dates,
        '_sea_coef_knot': ktrlite._aggregate_posteriors['map']['coef_knot'],
        '_seasonality': ktrlite._seasonality,
        '_seasonality_fs_order': ktrlite._seasonality_fs_order,
    }
    ktrx = KTRXAggregated(
        response_col='response',
        date_col='date',
        regressor_col=['a', 'b', 'c'],
        coef_prior_list=coef_prior_list,
        level_knot_dates=level_knot_dates,
        level_knots=level_knots,
        seasonal_knots_input=seasonal_knots_input,
        estimator_type=PyroEstimatorVI,
    )

    ktrx.fit(train_df)
    predict_df = ktrx.predict(test_df)

    expected_columns = ['date', 'prediction_5', 'prediction', 'prediction_95']
    expected_shape = (364, len(expected_columns))
    expected_num_parameters = 7

    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns
    assert len(ktrx._posterior_samples) == expected_num_parameters
    assert smape(test_df['response'].values, predict_df['prediction'].values) <= 0.5



from __future__ import absolute_import, division, print_function
from pystan import StanModel
import pickle
import pkg_resources
import os


def compile_stan_model(stan_model_name):
    """
    Compile stan model and save as pkl
    """
    source_model = pkg_resources.resource_filename(
        'orbit',
        'stan/{}.stan'.format(stan_model_name)
    )
    compiled_model = pkg_resources.resource_filename(
        'orbit',
        'stan_compiled/{}.pkl'.format(stan_model_name)
    )

    # updated for py3
    os.makedirs(os.path.dirname(compiled_model), exist_ok=True)

    # compile if stan source has changed
    if not os.path.isfile(compiled_model) or \
            os.path.getmtime(compiled_model) < os.path.getmtime(source_model):

        with open(source_model, encoding="utf-8") as f:
            model_code = f.read()

        sm = StanModel(model_code=model_code)

        with open(compiled_model, 'wb') as f:
            pickle.dump(sm, f, protocol=pickle.HIGHEST_PROTOCOL)

    return None


def get_compiled_stan_model(stan_model_name):
    """
    Load compiled Stan model
    """

    compile_stan_model(stan_model_name)

    model_file = pkg_resources.resource_filename(
        'orbit',
        'stan_compiled/{}.pkl'.format(stan_model_name)
    )
    with open(model_file, 'rb') as f:
        return pickle.load(f)



import pytest
import pandas as pd
import numpy as np

from orbit.diagnostics.backtest import TimeSeriesSplitter, BackTester
from orbit.diagnostics.metrics import smape, wmape, mape, mse, mae, rmsse
from orbit.models.lgt import LGTMAP
from orbit.models.ktrlite import KTRLiteMAP

@pytest.mark.parametrize(
    "scheduler_args",
    [
        {
            'min_train_len': 100,
            'incremental_len': 100,
            'forecast_len': 20
        },
        {
            'incremental_len': 100,
            'forecast_len': 20,
            'n_splits': 3
        },
        {
            'forecast_len': 20,
            'n_splits': 1
        },
        {
            'forecast_len': 1,
            'n_splits': 3
        }
    ],
    ids=['use_min_train_len', 'use_n_splits', 'single_split', 'one_forecast_len']
)
def test_time_series_splitter(scheduler_args):
    np_array = np.random.randn(300, 4)
    df = pd.DataFrame(np_array)

    tss = TimeSeriesSplitter(
        df=df,
        **scheduler_args,
    )

    assert tss.n_splits == len(list(tss.split()))


@pytest.mark.parametrize(
    "scheduler_args",
    [
        {
            'min_train_len': 100,
            'incremental_len': 100,
            'forecast_len': 20
        },
        {
            'incremental_len': 100,
            'forecast_len': 20,
            'n_splits': 3
        },
        {
            'forecast_len': 20,
            'n_splits': 1
        },
        {
            'forecast_len': 1,
            'n_splits': 3
        }
    ],
    ids=['use_min_train_len', 'use_n_splits', 'single_split', 'one_forecast_len']
)
def test_backtester_sceduler_args(iclaims_training_data, scheduler_args):
    df = iclaims_training_data

    lgt = LGTMAP(
        response_col='claims',
        date_col='week',
        seasonality=1,
        verbose=False
    )

    backtester = BackTester(
        model=lgt,
        df=df,
        **scheduler_args,
    )

    backtester.fit_predict()
    eval_out = backtester.score(metrics=[smape])
    assert np.all(eval_out['metric_values'].values > 0)


@pytest.mark.parametrize(
    "metrics",
    [None, [smape, wmape, mae], smape],
    ids=['default', 'multi-metrics', 'single-metric']
)
def test_backtester_test_metrics(iclaims_training_data, metrics):
    df = iclaims_training_data

    lgt = LGTMAP(
        response_col='claims',
        date_col='week',
        seasonality=1,
        verbose=False
    )

    backtester = BackTester(
        model=lgt,
        df=df,
        forecast_len=3,
        n_splits=1,
    )

    backtester.fit_predict()
    eval_out = backtester.score(metrics=metrics)
    evaluated_metrics = set(eval_out['metric_name'].tolist())

    if metrics is None:
        expected_metrics = [x.__name__ for x in backtester._default_metrics]
    elif isinstance(metrics, list):
        expected_metrics = [x.__name__ for x in metrics]
    else:
        expected_metrics = [metrics.__name__]

    assert set(expected_metrics) == evaluated_metrics

@pytest.mark.parametrize(
    "missing_flag",
    [False, True],
    ids=['full-values', 'missing-values']
)
def test_backtester_ktr_and_missing_val(make_daily_data, missing_flag):
    train_df, test_df, _ = make_daily_data
    df = pd.concat([train_df, test_df], axis=0, ignore_index=True)
    if missing_flag:
        # create a missing value in testing
        df.loc[df.shape[0] - 3, 'response'] = np.nan
        # create a missing value in training
        df.loc[10, 'response'] = np.nan

    ktr = KTRLiteMAP(
        date_col='date',
        response_col='response',
        seasonality=[365.25],
        verbose=False
    )

    bt = BackTester(
        model=ktr,
        df=df,
        n_splits=3,
        incremental_len=100,
        forecast_len=20,
    )

    bt.fit_predict()
    predicted_df = bt.get_predicted_df()
    assert set(predicted_df['split_key'].tolist()) == {0, 1, 2}

    bt_score_df = bt.score(include_training_metrics=False)
    num_testing_metrics = 6
    expected_shape = (num_testing_metrics, 3)
    assert bt_score_df.shape == expected_shape

    testing_metrics_df = bt_score_df[~bt_score_df['is_training_metric']]
    # rmsse is the only one not working for null values; otherwise, they should have valid values
    if missing_flag:
        metric_vals = testing_metrics_df.loc[
            testing_metrics_df['metric_name'] != 'rmsse', 'metric_values'].values
        assert np.all(~np.isnan(metric_vals))
        missing_metric_val = testing_metrics_df.loc[
            testing_metrics_df['metric_name'] == 'rmsse', 'metric_values'].values
        assert np.all(np.isnan(missing_metric_val))
    else:
        metric_vals = testing_metrics_df['metric_values'].values
        assert np.all(~np.isnan(metric_vals))


def test_backtester_with_training_data(iclaims_training_data):
    df = iclaims_training_data

    lgt = LGTMAP(
        response_col='claims',
        date_col='week',
        seasonality=1,
        verbose=False
    )

    backtester = BackTester(
        model=lgt,
        df=df,
        min_train_len=100,
        incremental_len=100,
        forecast_len=20,
    )

    backtester.fit_predict()
    eval_out = backtester.score(include_training_metrics=True)
    evaluated_test_metrics = set(eval_out.loc[~eval_out['is_training_metric'], 'metric_name'].tolist())
    evaluated_train_metrics = set(eval_out.loc[eval_out['is_training_metric'], 'metric_name'].tolist())

    expected_test_metrics = [x.__name__ for x in backtester._default_metrics]

    expected_train_metrics = list(filter(
        lambda x: backtester._get_metric_callable_signature(x) == {'actual', 'predicted'}, backtester._default_metrics)
    )
    expected_train_metrics = [x.__name__ for x in expected_train_metrics]

    assert set(expected_test_metrics) == evaluated_test_metrics
    assert set(expected_train_metrics) == evaluated_train_metrics

    # default metric has 6 values where rmsse is only used in test metric
    num_training_metrics = 5
    num_testing_metrics = 6

    train_metric_val = eval_out.loc[eval_out['is_training_metric'], 'metric_values'].values
    test_metric_val = eval_out.loc[~eval_out['is_training_metric'], 'metric_values'].values

    assert len(train_metric_val) == num_training_metrics
    assert len(test_metric_val) == num_testing_metrics
    assert np.all(~np.isnan(train_metric_val))
    assert np.all(~np.isnan(test_metric_val))



import custom_inherit as ci
from copy import copy, deepcopy
import numpy as np
import pandas as pd

from ..constants.constants import PredictMethod, PredictionKeys
from ..estimators.stan_estimator import StanEstimatorMCMC
from ..utils.docstring_style import merge_numpy_docs_dedup
from ..utils.predictions import prepend_date_column, compute_percentiles
from ..exceptions import IllegalArgument, ModelException, PredictionException, AbstractMethodException
from ..utils.general import is_ordered_datetime


ci.store["numpy_with_merge_dedup"] = merge_numpy_docs_dedup
ci.add_style("numpy_with_merge_dedup", merge_numpy_docs_dedup)


class BaseTemplate(object, metaclass=ci.DocInheritMeta(style="numpy_with_merge_dedup")):
    """ Base abstract class for univariate time-series model creation
    `BaseTemplate` will instantiate an estimator class of `estimator_type`.
    Each model defines its own `_supported_estimator_types` to determine if
    the provided `estimator_type` is supported for that particular model.
    Parameters
    ----------
    response_col : str
        Name of response variable column, default 'y'
    date_col : str
        Name of date variable column, default 'ds'
    estimator_type : orbit.BaseEstimator
        Any subclass of `orbit.BaseEstimator`
    Notes
    -----
    For attributes which are input by users and needed to mutate further downstream, we will introduce a
    new internal attribute with identical name except a prefix "_".
    e.g. If x appear in the arg default as `None` and we need to impute by 0. we will have self._x = 0 downstream.
    """
    # TODO: for now we assume has to be ENUM, maybe we can allow list as well? such that left and right are always
    # using same name
    # data labels for sampler
    _data_input_mapper = None
    # used to match name of `*.stan` or `*.pyro` file to look for the model
    _model_name = None
    # TODO: right now we assume _fitter is only for one specific estimator type
    # TODO: in the future, we should make it for example like a dict {PyroEstimator: pyro_model} etc. in case we want
    # TODO: to support multiple estimators and use for validation
    # EXPERIMENTAL: _fitter is used for quick supply of pyro / stan object instead of supplying a file
    _fitter = None
    # supported estimators in ..estimators
    # concrete classes should overwrite this
    _supported_estimator_types = None  # set for each model

    def __init__(self, response_col='y', date_col='ds', estimator_type=StanEstimatorMCMC, **kwargs):
        # general fields passed into Base Template
        self.response_col = response_col
        self.date_col = date_col

        # basic response fields
        # mainly set by ._set_training_df_meta() and ._set_dynamic_attributes()
        self.response = None
        self.date_array = None
        self.num_of_observations = None
        self.training_start = None
        self.training_end = None
        self._model_data_input = None

        # basic estimator fields
        self.estimator_type = estimator_type
        self.estimator = self.estimator_type(**kwargs)
        self.with_mcmc = None
        # set by ._set_init_values
        # this is ONLY used by stan which by default used 'random'
        self._init_values = None

        self._validate_supported_estimator_type()
        self._set_with_mcmc()

        # set by _set_model_param_names()
        self._model_param_names = list()

        # set by `fit()`
        self._posterior_samples = dict()
        # init aggregate posteriors
        self._aggregate_posteriors = dict()

        # for full Bayesian, user can store full prediction array if requested
        self.prediction_array = None
        self.prediction_input_meta = dict()

        # storing metrics in training result meta
        self._training_metrics = dict()

    # initialization related modules
    def _validate_supported_estimator_type(self):
        if self.estimator_type not in self._supported_estimator_types:
            msg_template = "Model class: {} is incompatible with Estimator: {}.  Estimator Support: {}"
            model_class = type(self)
            estimator_type = self.estimator_type
            raise IllegalArgument(
                msg_template.format(model_class, estimator_type, str(self._supported_estimator_types))
            )

    def _set_with_mcmc(self):
        """Include extra indicator to indicate whether the object is using mcmc type of estimator
        """
        estimator_type = self.estimator_type
        # set `with_mcmc` attribute based on estimator type
        # if no attribute for _is_mcmc_estimator, default to False
        if getattr(estimator_type, '_is_mcmc_estimator', False):
            self.with_mcmc = 1
        else:
            self.with_mcmc = 0

    def _set_model_param_names(self, **kwargs):
        """Set label for model parameters.  This function can be dependent on
        static attributes.
        """
        raise AbstractMethodException(
            "Abstract method.  Model should implement concrete ._set_model_param_names().")

    def get_model_param_names(self):
        return self._model_param_names

    def _set_static_attributes(self, **kwargs):
        """Set static attributes which are independent from data matrix.
        These methods are supposed to be over-ride by child (model) template.
        For attributes dependent on data matrix, use _set_dynamic_attributes
        """
        pass

    # fit and predict related modules
    def _validate_training_df(self, df):
        df_columns = df.columns

        # validate date_col
        if self.date_col not in df_columns:
            raise ModelException("DataFrame does not contain `date_col`: {}".format(self.date_col))

        # validate ordering of time series
        date_array = pd.to_datetime(df[self.date_col]).reset_index(drop=True)
        if not is_ordered_datetime(date_array):
            raise ModelException('Datetime index must be ordered and not repeat')

        # validate response variable is in df
        if self.response_col not in df_columns:
            raise ModelException("DataFrame does not contain `response_col`: {}".format(self.response_col))

    def _set_training_df_meta(self, df):
        self.response = df[self.response_col].values
        self.date_array = pd.to_datetime(df[self.date_col]).reset_index(drop=True)
        self.num_of_observations = len(self.response)
        self.response_sd = np.nanstd(self.response)
        self.training_start = df[self.date_col].iloc[0]
        self.training_end = df[self.date_col].iloc[-1]

    def _set_model_data_input(self):
        """Collects data attributes into a dict for sampling/optimization api"""
        # refresh a clean dict
        data_inputs = dict()

        if not self._data_input_mapper:
            raise ModelException('Empty or invalid data_input_mapper')

        for key in self._data_input_mapper:
            # mapper keys in upper case; inputs in lower case
            key_lower = key.name.lower()
            input_value = getattr(self, key_lower, None)
            if input_value is None:
                raise ModelException('{} is missing from data input'.format(key_lower))
            if isinstance(input_value, bool):
                # stan accepts bool as int only
                input_value = int(input_value)
            data_inputs[key.value] = input_value

        self._model_data_input = data_inputs

    def get_model_data_input(self):
        return self._model_data_input

    def _set_init_values(self):
        """Set init as a callable (for Stan ONLY)
        See: https://pystan.readthedocs.io/en/latest/api.htm
        """
        pass

    def get_init_values(self):
        return self._init_values

    def is_fitted(self):
        # if either aggregate posterior and posterior_samples are non-empty, claim it as fitted model (true),
        # else false.
        if bool(self._posterior_samples):
            return True
        for key in self._aggregate_posteriors.keys():
            if bool(self._aggregate_posteriors[key]):
                return True
        return False

    def _set_dynamic_attributes(self, df):
        """Set required input based on input DataFrame, rather than at object instantiation"""
        pass

    def fit(self, df):
        """Fit model to data and set extracted posterior samples"""
        estimator = self.estimator
        model_name = self._model_name
        df = df.copy()

        # default set and validation of input data frame
        self._validate_training_df(df)
        self._set_training_df_meta(df)

        # customize module
        self._set_dynamic_attributes(df)

        # default process post attributes setting
        # _set_model_data_input() behavior depends on _set_training_df_meta()
        self._set_model_data_input()
        # set initial values for randomization; right now only used by pystan; default as 'random'
        self._set_init_values()

        # estimator inputs
        data_input = self.get_model_data_input()
        init_values = self.get_init_values()
        model_param_names = self.get_model_param_names()

        # note that estimator will search for the .stan, .pyro model file based on the
        # estimator type and model_name provided
        model_extract, training_metrics = estimator.fit(
            model_name=model_name,
            model_param_names=model_param_names,
            data_input=data_input,
            fitter=self._fitter,
            init_values=init_values
        )

        self._posterior_samples = model_extract
        self._training_metrics = training_metrics

    def get_prediction_df_meta(self, df):
        # get prediction df meta
        prediction_df_meta = {
            'date_array': pd.to_datetime(df[self.date_col]).reset_index(drop=True),
            'df_length': len(df.index),
            'prediction_start': df[self.date_col].iloc[0],
            'prediction_end': df[self.date_col].iloc[-1]
        }

        if not is_ordered_datetime(prediction_df_meta['date_array']):
            raise IllegalArgument('Datetime index must be ordered and not repeat')

        # TODO: validate that all regressor columns are present, if any

        if prediction_df_meta['prediction_start'] < self.training_start:
            raise PredictionException('Prediction start must be after training start.')

        return prediction_df_meta

    def get_posterior_samples(self):
        return self._posterior_samples.copy()

    def get_training_metrics(self):
        return self._training_metrics.copy()

    def get_prediction_input_meta(self, df):
        # remove reference from original input
        df = df.copy()

        # get prediction df meta
        prediction_input_meta = {
            'date_array': pd.to_datetime(df[self.date_col]).reset_index(drop=True),
            'df_length': len(df.index),
            'prediction_start': df[self.date_col].iloc[0],
            'prediction_end': df[self.date_col].iloc[-1],
        }

        if not is_ordered_datetime(prediction_input_meta['date_array']):
            raise IllegalArgument('Datetime index must be ordered and not repeat')

        # TODO: validate that all regressor columns are present, if any

        if prediction_input_meta['prediction_start'] < self.training_start:
            raise PredictionException('Prediction start must be after training start.')

        trained_len = self.num_of_observations

        # If we cannot find a match of prediction range, assume prediction starts right after train
        # end
        if prediction_input_meta['prediction_start'] > self.training_end:
            forecast_dates = set(prediction_input_meta['date_array'])
            n_forecast_steps = len(forecast_dates)
            # time index for prediction start
            start = trained_len
        else:
            # compute how many steps to forecast
            forecast_dates = \
                set(prediction_input_meta['date_array']) - set(self.date_array)
            # check if prediction df is a subset of training df
            # e.g. "negative" forecast steps
            n_forecast_steps = len(forecast_dates) or - (
                len(set(self.date_array) - set(prediction_input_meta['date_array']))
            )
            # time index for prediction start
            start = pd.Index(
                self.date_array).get_loc(prediction_input_meta['prediction_start'])

        prediction_input_meta.update({
            'start': start,
            'n_forecast_steps': n_forecast_steps,
        })

        self.prediction_input_meta = prediction_input_meta

    def predict(self, df, **kwargs):
        """Predict interface for users"""
        raise AbstractMethodException("Abstract method.  Model should implement concrete .predict().")

    def _predict(self, posterior_estimates, df, include_error=False, **kwargs):
        """Inner predict being called internal for different prediction purpose such as bootstrapping"""
        raise AbstractMethodException("Abstract method.  Model should implement concrete ._predict().")


class MAPTemplate(BaseTemplate):
    """ Abstract class for MAP (Maximum a Posteriori) prediction
    In this module, prediction is based on Maximum a Posteriori (aka Mode) of the posterior.
    This template only supports MAP inference.
    Parameters
    ----------
    n_bootstrap_draws : int
        Number of bootstrap samples to draw from the error part to generate the uncertainty.
        If set to be -1, will use the original posterior draw (no uncertainty).
    prediction_percentiles : list
        List of integers of prediction percentiles that should be returned on prediction. To avoid reporting any
        confident intervals, pass an empty list
    """

    def __init__(self, n_bootstrap_draws=1e4, prediction_percentiles=None, **kwargs):
        super().__init__(**kwargs)

        # n_bootstrap_draws here only to provide empirical prediction percentiles;
        # mid-point estimate is always replaced
        self.n_bootstrap_draws = n_bootstrap_draws
        self.prediction_percentiles = prediction_percentiles
        self._prediction_percentiles = None

        # unlike full prediction, it does not take negative number of bootstrap draw
        # if self.n_bootstrap_draws < 2:
        #     raise IllegalArgument("Error: Number of bootstrap draws must be at least 2")
        if self.prediction_percentiles is None:
            self._prediction_percentiles = [5, 95]
        else:
            self._prediction_percentiles = copy(self.prediction_percentiles)

        self._prediction_percentiles += [50]  # always find median
        self._prediction_percentiles = list(set(self._prediction_percentiles))  # unique set
        self._prediction_percentiles.sort()

        # override init aggregate posteriors
        self._aggregate_posteriors = {PredictMethod.MAP.value: dict()}

        self._set_static_attributes()
        self._set_model_param_names()

    # fit and predict related modules
    def _set_map_posterior(self):
        """ set MAP posteriors with right dimension"""
        posterior_samples = self._posterior_samples
        map_posterior = {}
        for param_name in self._model_param_names:
            param_array = posterior_samples[param_name]
            # add dimension so it works with vector math in `_predict`
            param_array = np.expand_dims(param_array, axis=0)
            map_posterior.update({param_name: param_array})

        self._aggregate_posteriors[PredictMethod.MAP.value] = map_posterior

    def fit(self, df):
        super().fit(df)
        self._set_map_posterior()

    def predict(self, df, decompose=False, **kwargs):
        # raise if model is not fitted
        if not self.is_fitted():
            raise PredictionException("Model is not fitted yet.")
        # obtain basic meta data from input df
        self.get_prediction_input_meta(df)

        # perform point prediction
        aggregate_posteriors = self._aggregate_posteriors.get(PredictMethod.MAP.value)
        point_predicted_dict = self._predict(
            posterior_estimates=aggregate_posteriors, df=df, include_error=False, **kwargs
        )
        for k, v in point_predicted_dict.items():
            point_predicted_dict[k] = np.squeeze(v, 0)

        # to derive confidence interval; the condition should be sufficient since we add [50] by default
        if self.n_bootstrap_draws > 0 and len(self._prediction_percentiles) > 1:
            # perform bootstrap; we don't have posterior samples. hence, we just repeat the draw here.
            posterior_samples = {}
            for k, v in aggregate_posteriors.items():
                posterior_samples[k] = np.repeat(v, self.n_bootstrap_draws, axis=0)
            predicted_dict = self._predict(
                posterior_estimates=posterior_samples, df=df, include_error=True, **kwargs
            )
            percentiles_dict = compute_percentiles(predicted_dict, self._prediction_percentiles)
            # replace mid point prediction by point estimate
            percentiles_dict.update(point_predicted_dict)

            if PredictionKeys.PREDICTION.value not in percentiles_dict.keys():
                raise PredictionException("cannot find the key:'{}' from return of _predict()".format(
                    PredictionKeys.PREDICTION.value))

            # reduce to prediction only if decompose is not requested
            if not decompose:
                k = PredictionKeys.PREDICTION.value
                reduced_keys = [k + "_" + str(p) if p != 50 else k for p in self._prediction_percentiles]
                percentiles_dict = {k: v for k, v in percentiles_dict.items() if k in reduced_keys}
            predicted_df = pd.DataFrame(percentiles_dict)
        else:
            # reduce to prediction only if decompose is not requested
            if not decompose:
                point_predicted_dict = {
                    k: v for k, v in point_predicted_dict.items() if k == PredictionKeys.PREDICTION.value
                }
            predicted_df = pd.DataFrame(point_predicted_dict)

        predicted_df = prepend_date_column(predicted_df, df, self.date_col)
        return predicted_df


class FullBayesianTemplate(BaseTemplate):
    """ Abstract class for full Bayesian prediction
    In full prediction, the prediction occurs as a function of each parameter posterior sample,
    and the prediction results are aggregated after prediction. Prediction will
    always return the median (aka 50th percentile) along with any additional percentiles that
    are specified.
    Parameters
    ----------
    n_bootstrap_draws : int
        Number of bootstrap samples to draw from the initial MCMC or VI posterior samples.
        If -1, use the original posterior draws.
    prediction_percentiles : list
        List of integers of prediction percentiles that should be returned on prediction. To avoid reporting any
        confident intervals, pass an empty list
    """
    def __init__(self, n_bootstrap_draws=-1, prediction_percentiles=None, **kwargs):
        super().__init__(**kwargs)
        self.n_bootstrap_draws = n_bootstrap_draws
        self.prediction_percentiles = prediction_percentiles

        # set default args
        self._prediction_percentiles = None
        self._n_bootstrap_draws = self.n_bootstrap_draws

        if self.prediction_percentiles is None:
            self._prediction_percentiles = [5, 95]
        else:
            self._prediction_percentiles = copy(self.prediction_percentiles)

        self._prediction_percentiles += [50]  # always find median
        self._prediction_percentiles = list(set(self._prediction_percentiles))  # unique set
        self._prediction_percentiles.sort()

        if not self.n_bootstrap_draws:
            self._n_bootstrap_draws = -1

        # init aggregate posteriors
        self._aggregate_posteriors = {
            PredictMethod.MEAN.value: dict(),
            PredictMethod.MEDIAN.value: dict(),
        }

        self._set_static_attributes()
        self._set_model_param_names()

    @staticmethod
    def _bootstrap(num_samples, posterior_samples, n):
        """Draw `n` number of bootstrap samples from the posterior_samples.
        Args
        ----
        n : int
            The number of bootstrap samples to draw
        """
        if n < 2:
            raise IllegalArgument("Error: Number of bootstrap draws must be at least 2")

        sample_idx = np.random.choice(range(num_samples), size=n, replace=True)
        bootstrap_samples_dict = {}
        for k, v in posterior_samples.items():
            bootstrap_samples_dict[k] = v[sample_idx]

        return bootstrap_samples_dict

    def _set_aggregate_posteriors(self):
        posterior_samples = self._posterior_samples
        mean_posteriors = {}
        median_posteriors = {}

        # for each model param, aggregate using `method`
        for param_name in self._model_param_names:
            param_ndarray = posterior_samples[param_name]

            mean_posteriors.update(
                {param_name: np.mean(param_ndarray, axis=0, keepdims=True)},
            )

            median_posteriors.update(
                {param_name: np.median(param_ndarray, axis=0, keepdims=True)},
            )

        self._aggregate_posteriors[PredictMethod.MEAN.value] = mean_posteriors
        self._aggregate_posteriors[PredictMethod.MEDIAN.value] = median_posteriors

    def predict(self, df, decompose=False, store_prediction_array=False, **kwargs):
        # raise if model is not fitted
        if not self.is_fitted():
            raise PredictionException("Model is not fitted yet.")
        # obtain basic meta data from input df
        self.get_prediction_input_meta(df)

        # if bootstrap draws, replace posterior samples with bootstrap
        posterior_samples = self._bootstrap(
            num_samples=self.estimator.num_sample,
            posterior_samples=self._posterior_samples,
            n=self._n_bootstrap_draws
        ) if self._n_bootstrap_draws > 1 else self._posterior_samples

        predicted_dict = self._predict(
            posterior_estimates=posterior_samples, df=df, include_error=True, **kwargs
        )

        if PredictionKeys.PREDICTION.value not in predicted_dict.keys():
            raise PredictionException("cannot find the key:'{}' from return of _predict()".format(
                PredictionKeys.PREDICTION.value))

        # reduce to prediction only if decompose is not requested
        # note that unlike other template, we can filter the keys before percentiles computation
        # hence, we can reduce with mapping on PredictionKeys
        if not decompose:
            predicted_dict = {k: v for k, v in predicted_dict.items() if k == PredictionKeys.PREDICTION.value}

        if store_prediction_array:
            self.prediction_array = predicted_dict[PredictionKeys.PREDICTION.value]
        percentiles_dict = compute_percentiles(predicted_dict, self._prediction_percentiles)
        predicted_df = pd.DataFrame(percentiles_dict)
        predicted_df = prepend_date_column(predicted_df, df, self.date_col)
        return predicted_df


class AggregatedPosteriorTemplate(BaseTemplate):
    """ Abstract class for full aggregated posteriors prediction
    In aggregated prediction, the parameter posterior samples are reduced using `aggregate_method`
    before performing a single prediction. This template only supports aggregated posterior inference.
    Parameters
    ----------
    aggregate_method : { 'mean', 'median' }
        Method used to reduce parameter posterior samples
    n_bootstrap_draws : int
        Number of bootstrap samples to draw from the error part to generate the uncertainty.
        If -1, will use the original posterior draw (no uncertainty).
    prediction_percentiles : list
        List of integers of prediction percentiles that should be returned on prediction. To avoid reporting any
        confident intervals, pass an empty list
    """
    def __init__(self, aggregate_method='mean', n_bootstrap_draws=1e4, prediction_percentiles=None, **kwargs):
        super().__init__(**kwargs)
        # n_bootstrap_draws here only to provide empirical prediction percentiles;
        # mid-point estimate is always replaced
        self.n_bootstrap_draws = n_bootstrap_draws
        self.prediction_percentiles = prediction_percentiles
        self._prediction_percentiles = None
        if self.prediction_percentiles is None:
            self._prediction_percentiles = [5, 95]
        else:
            self._prediction_percentiles = copy(self.prediction_percentiles)

        self._prediction_percentiles += [50]  # always find median
        self._prediction_percentiles = list(set(self._prediction_percentiles ))  # unique set
        self._prediction_percentiles .sort()
        # unlike full prediction, it does not take negative number of bootstrap draw
        # if self.n_bootstrap_draws < 2:
        #     raise IllegalArgument("Error: Number of bootstrap draws must be at least 2")

        self.aggregate_method = aggregate_method
        # override init aggregate posteriors
        self._aggregate_posteriors = {aggregate_method: dict()}
        self._validate_aggregate_method()

        # init aggregate posteriors
        self._aggregate_posteriors = {
            PredictMethod.MEAN.value: dict(),
            PredictMethod.MEDIAN.value: dict(),
        }

        self._set_static_attributes()
        self._set_model_param_names()

    def _validate_aggregate_method(self):
        if self.aggregate_method not in list(self._aggregate_posteriors.keys()):
            raise PredictionException("No aggregate method defined for: `{}`".format(self.aggregate_method))

    def _set_aggregate_posteriors(self):
        posterior_samples = self._posterior_samples
        mean_posteriors = {}
        median_posteriors = {}

        # for each model param, aggregate using `method`
        for param_name in self._model_param_names:
            param_ndarray = posterior_samples[param_name]

            mean_posteriors.update(
                {param_name: np.mean(param_ndarray, axis=0, keepdims=True)},
            )

            median_posteriors.update(
                {param_name: np.median(param_ndarray, axis=0, keepdims=True)},
            )

        self._aggregate_posteriors[PredictMethod.MEAN.value] = mean_posteriors
        self._aggregate_posteriors[PredictMethod.MEDIAN.value] = median_posteriors

    def fit(self, df, keep_posterior_samples=True):
        super().fit(df)
        self._set_aggregate_posteriors()
        if not keep_posterior_samples:
            self._posterior_samples = {}

    def predict(self, df, decompose=False, **kwargs):
        # raise if model is not fitted
        if not self.is_fitted():
            raise PredictionException("Model is not fitted yet.")
        # obtain basic meta data from input df
        self.get_prediction_input_meta(df)

        # perform point prediction
        aggregate_posteriors = self._aggregate_posteriors.get(self.aggregate_method)
        point_predicted_dict = self._predict(
            posterior_estimates=aggregate_posteriors, df=df, include_error=False, **kwargs
        )
        for k, v in point_predicted_dict.items():
            point_predicted_dict[k] = np.squeeze(v, 0)

        # to derive confidence interval; the condition should be sufficient since we add [50] by default
        if self.n_bootstrap_draws > 0 and len(self._prediction_percentiles) > 1:
            # perform bootstrap; we don't have posterior samples. hence, we just repeat the draw here.
            posterior_samples = {}
            for k, v in aggregate_posteriors.items():
                posterior_samples[k] = np.repeat(v, self.n_bootstrap_draws, axis=0)
            predicted_dict = self._predict(
                posterior_estimates=posterior_samples, df=df, include_error=True, **kwargs
            )
            percentiles_dict = compute_percentiles(predicted_dict, self._prediction_percentiles)
            # replace mid point prediction by point estimate
            percentiles_dict.update(point_predicted_dict)

            if PredictionKeys.PREDICTION.value not in percentiles_dict.keys():
                raise PredictionException("cannot find the key:'{}' from return of _predict()".format(
                    PredictionKeys.PREDICTION.value))

            # reduce to prediction only if decompose is not requested
            if not decompose:
                k = PredictionKeys.PREDICTION.value
                reduced_keys = [k + "_" + str(p) if p != 50 else k for p in self._prediction_percentiles]
                percentiles_dict = {k: v for k, v in percentiles_dict.items() if k in reduced_keys}
            predicted_df = pd.DataFrame(percentiles_dict)
        else:
            # reduce to prediction only if decompose is not requested
            if not decompose:
                point_predicted_dict = {
                    k: v for k, v in point_predicted_dict.items() if k == PredictionKeys.PREDICTION.value
                }
            predicted_df = pd.DataFrame(point_predicted_dict)

        predicted_df = prepend_date_column(predicted_df, df, self.date_col)
        return predicted_df


from enum import Enum


class DataInputMapper(Enum):
    """
    mapping from object input to sampler
    """
    # ----------  Data Input ---------- #
    # observation related
    NUM_OF_OBSERVATIONS = 'NUM_OF_OBS'
    RESPONSE = 'RESPONSE'
    RESPONSE_SD = 'RESPONSE_SD'
    # ---------- Seasonality ---------- #
    _SEASONALITY = 'SEASONALITY'
    _SEASONALITY_SM_INPUT = 'SEA_SM_INPUT'
    # ---------- Common Local Trend ---------- #
    _LEVEL_SM_INPUT = 'LEV_SM_INPUT'
    WITH_MCMC = 'WITH_MCMC'


class BaseSamplingParameters(Enum):
    """
    base parameters in posteriors sampling
    """
    # ---------- Common Local Trend ---------- #
    LOCAL_TREND_LEVELS = 'l'
    LEVEL_SMOOTHING_FACTOR = 'lev_sm'
    # ---------- Noise Trend ---------- #
    RESIDUAL_SIGMA = 'obs_sigma'


class SeasonalitySamplingParameters(Enum):
    """
    seasonality component related parameters in posteriors sampling
    """
    SEASONALITY_LEVELS = 's'
    SEASONALITY_SMOOTHING_FACTOR = 'sea_sm'


class LatentSamplingParameters(Enum):
    """
    latent variables to be sampled
    """
    INITIAL_SEASONALITY = 'init_sea'



import numpy as np
EPS = 1e-5


def smape(actual, predicted):
    filtered = np.abs(actual) > EPS
    actual = actual[filtered]
    predicted = predicted[filtered]
    return 2 * np.mean(np.abs(actual - predicted) / (np.abs(actual) + np.abs(predicted)))


def mape(actual, predicted):
    filtered = np.abs(actual) > EPS
    actual = actual[filtered]
    predicted = predicted[filtered]
    return np.mean(np.abs((actual - predicted) / actual))


def wmape(actual, predicted):
    filtered = np.abs(actual) > EPS
    actual = actual[filtered]
    predicted = predicted[filtered]
    weights = np.abs(actual) / np.sum(np.abs(actual))
    return np.sum(weights * np.abs((actual - predicted) / actual))


def mae(actual, predicted):
    filtered = ~np.isnan(actual)
    actual = actual[filtered]
    predicted = predicted[filtered]
    return np.mean(np.abs(actual - predicted))


def mse(actual, predicted):
    filtered = ~np.isnan(actual)
    actual = actual[filtered]
    predicted = predicted[filtered]
    return np.mean(np.square(actual - predicted))


def rmsse(test_actual, test_predicted, train_actual):
    """Computes Root Mean Squared Scaled Error (RMSSE)

    A variant of the well-known Mean Absolute Scaled Error (MASE) proposed by Hyndman and Koehler (2006)

    Intuitively, the general idea is to measure additional value added by the model
    compared to a naive lag-1 predictor.

    .. math::
        \sqrt{\frac{1}{h}\frac{\sum^{n+h}_{t=n+1}(Y_t-\hat{Y}_t)^2}{\frac{1}{n-1}\sum^{n}_{t=2}{(Y_t-Y_{t-1})^2}

    Notes
    -----
    Reference from https://mofc.unic.ac.cy/m5-guidelines/

    """

    # exclude leading zeros for training
    first_nz = np.min(np.flatnonzero(train_actual))
    train_actual = test_actual[first_nz:]

    lag1_mse = mse(train_actual[1:], train_actual[:-1])

    forecast_mse = mse(test_actual, test_predicted)

    return np.sqrt(forecast_mse / lag1_mse)



import numpy as np


def reduce_by_max(x, n=2):
    out = x.copy()
    out[np.argsort(x)[:-n]] = 0
    return out


# Gaussian-Kernel
# https://en.wikipedia.org/wiki/Kernel_smoother
def gauss_kernel(x, x_i, rho=1.0, alpha=1.0, n_reduce=-1, point_to_flatten=1):
    """
    Parameters
    ----------
    x : array-like
        points required to compute kernel weight
    x_i : array-like
        reference points location used to compute correspondent distance of each entry points
    rho : float
        smoothing parameter known as "length-scale" in gaussian process
    alpha : float
        marginal standard deviation parameter in gaussian process; one should use 1 in kernel regression
    n_reduce : int
        if greater 0 (default=-1), reduce number of positive weights to such input
    point_to_flatten : float
        the time point starting to flatten the weights; default is 1 for normalized time points

    Returns
    -------
    np.ndarray
        2D array with size N x M such that
        N as the number of entry points
        M as the number of reference points
        matrix entries hold the value of weight of each element

    See Also
    --------
      1. https://mc-stan.org/docs/2_24/stan-users-guide/gaussian-process-regression.html
      2. https://en.wikipedia.org/wiki/Local_regression
    """
    N = len(x)
    M = len(x_i)
    k = np.zeros((N, M), np.double)
    alpha_sq = alpha ** 2
    rho_sq_t2 = 2 * rho ** 2
    for n in range(N):
        if x[n] <= point_to_flatten:
            k[n, :] = alpha_sq * np.exp(-1 * (x[n] - x_i) ** 2 / rho_sq_t2)
        else:
            # last weights carried forward for future time points
            k[n, :] = alpha_sq * np.exp(-1 * (point_to_flatten - x_i) ** 2 / rho_sq_t2)

    if n_reduce > 0:
       k = np.apply_along_axis(reduce_by_max, axis=1, arr=k, n=n_reduce)

    k = k / np.sum(k, axis=1, keepdims=True)

    return k


def sandwich_kernel(x, x_i):
    """
    Parameters
    ----------
    x : array-like
        points required to compute kernel weight
    x_i : array-like
        reference points location used to compute correspondent distance of each entry points

    Returns
    -------
    np.ndarray
        2D array with size N x M such that
        N as the number of entry points
        M as the number of reference points
        matrix entries hold the value of weight of each element

    See Also
    --------
      1. https://mc-stan.org/docs/2_24/stan-users-guide/gaussian-process-regression.html
      2. https://en.wikipedia.org/wiki/Local_regression
    """
    N = len(x)
    M = len(x_i)
    k = np.zeros((N, M), dtype=np.double)

    np_idx = np.where(x < x_i[0])
    k[np_idx, 0] = 1

    for m in range(M - 1):
        np_idx = np.where(np.logical_and(x >= x_i[m], x < x_i[m + 1]))
        total_dist = x_i[m + 1] - x_i[m]
        backward_dist = x[np_idx] - x_i[m]
        forward_dist = x_i[m + 1] - x[np_idx]
        k[np_idx, m] = forward_dist / total_dist
        k[np_idx, m + 1] = backward_dist / total_dist

    np_idx = np.where(x >= x_i[M - 1])
    k[np_idx, M - 1] = 1

    # TODO: it is probably not needed
    k = k / np.sum(k, axis=1, keepdims=True)

    return k


def parabolic_kernel(x, x_i):
    N = len(x)
    M = len(x_i)
    k = np.zeros((N, M), dtype=np.double)

    # boundary case
    np_idx = np.where(x < x_i[0])
    if len(np_idx) > 0:
        k[np_idx, 0] = 1

    for m in range(M - 1):
        np_idx = np.where(np.logical_and(x >= x_i[m], x < x_i[m + 1]))
        total_dist = x_i[m + 1] - x_i[m]
        backward_dist = x[np_idx] - x_i[m]
        forward_dist = x_i[m + 1] - x[np_idx]
        k[np_idx, m] = 0.75 * (1 - (backward_dist / total_dist) ** 2)
        k[np_idx, m + 1] = 0.75 * (1 - (forward_dist / total_dist) ** 2)

    # boundary case
    np_idx = np.where(x >= x_i[M - 1])
    if len(np_idx) > 0:
        k[np_idx, M - 1] = 1

    # TODO: it is probably not needed
    k = k / np.sum(k, axis=1, keepdims=True)

    return k



import pandas as pd
import numpy as np


def load_iclaims(end_date='2018-06-24'):
    """Load iclaims dataset

    Returns
    -------
        pd.DataFrame

    Notes
    -----
    iclaims is a dataset containing the weekly initial claims for US unemployment benefits against a few related google
    trend queries (unemploy, filling and job)from Jan 2010 - June 2018. This aims to mimick the dataset from the paper
    Predicting the Present with Bayesian Structural Time Series by SCOTT and VARIAN (2014).
    Number of claims are obtained from [Federal Reserve Bank of St. Louis] while google queries are obtained through
    Google Trends API.

    Note that dataset is transformed by natural log before fitting in order to be fitted as a multiplicative model.

    https://fred.stlouisfed.org/series/ICNSA
    https://trends.google.com/trends/?geo=US
    https://finance.yahoo.com/
    """
    url = 'https://raw.githubusercontent.com/uber/orbit/master/examples/data/iclaims_example.csv'
    df = pd.read_csv(url, parse_dates=['week'])
    df = df[df['week'] <= end_date]

    # standardize the regressors by mean; equivalent to subtracting mean after np.log
    regressors = ['trend.unemploy', 'trend.filling', 'trend.job', 'sp500', 'vix']

    # convert to float
    for col in regressors:
        df[col] = df[col].astype(float)

    # log transfer
    df[['claims'] + regressors] = df[['claims'] + regressors].apply(np.log)
    # de-mean
    df[regressors] = df[regressors] - df[regressors].apply(np.mean)

    return df


def load_m4weekly():
    """Load m4 weekly sample dataset

    Returns
    -------
        pandas DataFrame

    Notes
    -----
    https://forecasters.org/resources/time-series-data/m4-competition/
    """
    url = 'https://raw.githubusercontent.com/uber/orbit/master/examples/data/m4_weekly.csv'
    df = pd.read_csv(url, parse_dates=['date'])

    return df


def load_m5daily():
    """Load m5 aggregated daily dataset

    Returns
    -------
        pandas DataFrame

    Notes
    -----
    https://www.kaggle.com/c/m5-forecasting-accuracy
    """
    url = 'https://raw.githubusercontent.com/uber/orbit/master/examples/data/m5_agg_demand_full.csv'
    df = pd.read_csv(url, parse_dates=['date'])

    return df


def load_m3monthly():
    """Load m3 monthly sample dataset

    Returns
    -------
        pandas DataFrame

    Notes
    -----
    https://forecasters.org/resources/time-series-data/m3-competition/
    """
    url = 'https://raw.githubusercontent.com/uber/orbit/master/examples/data/m3_monthly.csv'
    df = pd.read_csv(url, parse_dates=['date'])

    return df


def load_electricity_demand():
    """Load Turkish electricity demand daily data from 1 January 2000 to 31 December 2008.
    Returns
    -------
        pandas DataFrame

    Notes
    -----
    https://robjhyndman.com/publications/complex-seasonality/

    :return:
    """

    url = 'https://robjhyndman.com/data/turkey_elec.csv'
    df = pd.read_csv(url, header=None, names=['electricity'])
    df['date'] = pd.date_range(start='1/1/2000', end='31/12/2008', freq='D')
    # re-arrange columns
    df = df[['date', 'electricity']]

    return df


def load_air_passengers():
    """Load air passengers csv from prophet github
    Returns
    -------
        pandas DataFrame

    Notes
    -----
    https://github.com/facebook/prophet/tree/master/examples

    :return:
    """

    url = 'https://raw.githubusercontent.com/facebook/prophet/master/examples/example_air_passengers.csv'
    df = pd.read_csv(url, parse_dates=['ds'])

    return df



from abc import abstractmethod
import numpy as np
import pyro
from pyro.infer import SVI, Trace_ELBO
from pyro.infer.autoguide import AutoLowRankMultivariateNormal, AutoDelta
from pyro.optim import ClippedAdam

from .base_estimator import BaseEstimator
from ..exceptions import EstimatorException
from ..utils.pyro import get_pyro_model


class PyroEstimator(BaseEstimator):
    """Abstract PyroEstimator with shared args for all PyroEstimator child classes

    Parameters
    ----------
    num_steps : int
        Number of estimator steps in optimization
    learning_rate : float
        Estimator learning rate
    learning_rate_total_decay : float
        A config re-parameterized from ``lrd`` in :class:`~pyro.optim.ClippedAdam`. For example, 0.1 means a 90%
        reduction of the final step as of original learning rate where linear decay is implied along the steps. In the
        case of 1.0, no decay is applied.  All steps will have the constant learning rate specified by `learning_rate`.
    seed : int
        Seed int
    message : int
        Print to console every `message` number of steps
    kwargs
        Additional BaseEstimator args
    Notes
    -----
        See http://docs.pyro.ai/en/stable/_modules/pyro/optim/clipped_adam.html for optimizer details
    """
    def __init__(self, num_steps=1001, learning_rate=0.1, learning_rate_total_decay=1.0, message=100, **kwargs):
        super().__init__(**kwargs)
        self.num_steps = num_steps
        self.learning_rate = learning_rate
        self.learning_rate_total_decay = learning_rate_total_decay
        self.message = message

    @abstractmethod
    def fit(self, model_name, model_param_names, data_input, fitter=None, init_values=None):
        raise NotImplementedError('Concrete fit() method must be implemented')


class PyroEstimatorVI(PyroEstimator):
    """Pyro Estimator for VI Sampling

    Parameters
    ----------
    num_sample : int
        Number of samples ot draw for inference, default 100
    num_particles : int
        Number of particles used in :class: `~pyro.infer.Trace_ELBO` for SVI optimization
    init_scale : float
        Parameter used in `pyro.infer.autoguide`; recommend a larger number of small dataset
    kwargs
        Additional `PyroEstimator` class args

    """

    def __init__(self, num_sample=100, num_particles=100, init_scale=0.1, **kwargs):
        super().__init__(**kwargs)
        self.num_sample = num_sample
        self.num_particles = num_particles
        self.init_scale = init_scale

    def fit(self, model_name, model_param_names, data_input, fitter=None, init_values=None):
        # verbose is passed through from orbit.models.base_estimator
        verbose = self.verbose
        message = self.message
        learning_rate = self.learning_rate
        learning_rate_total_decay = self.learning_rate_total_decay
        num_sample = self.num_sample
        seed = self.seed
        num_steps = self.num_steps

        pyro.set_rng_seed(seed)
        if fitter is None:
            fitter = get_pyro_model(model_name)  # abstract
        model = fitter(data_input)  # concrete

        # Perform stochastic variational inference using an auto guide.
        pyro.clear_param_store()
        guide = AutoLowRankMultivariateNormal(model, init_scale=self.init_scale)
        optim = ClippedAdam({
            "lr": learning_rate,
            "lrd": learning_rate_total_decay ** (1 / num_steps)
        })
        elbo = Trace_ELBO(num_particles=self.num_particles, vectorize_particles=True)
        loss_elbo = list()
        svi = SVI(model, guide, optim, elbo)
        for step in range(num_steps):
            loss = svi.step()
            loss_elbo.append(loss)
            if verbose and step % message == 0:
                scale_rms = guide._loc_scale()[1].detach().pow(2).mean().sqrt().item()
                print("step {: >4d} loss = {:0.5g}, scale = {:0.5g}".format(step, loss, scale_rms))

        # Extract samples.
        vectorize = pyro.plate("samples", num_sample, dim=-1 - model.max_plate_nesting)
        with pyro.poutine.trace() as tr:
            samples = vectorize(guide)()
        with pyro.poutine.replay(trace=tr.trace):
            samples.update(vectorize(model)())

        # Convert from torch.Tensors to numpy.ndarrays.
        extract = {
            name: value.detach().squeeze().numpy()
            for name, value in samples.items()
        }

        # make sure that model param names are a subset of stan extract keys
        invalid_model_param = set(model_param_names) - set(list(extract.keys()))
        if invalid_model_param:
            raise EstimatorException("Pyro model definition does not contain required parameters")

        # `stan.optimizing` automatically returns all defined parameters
        # filter out unnecessary keys
        posteriors = {param: extract[param] for param in model_param_names}
        training_metrics = {'loss_elbo': np.array(loss_elbo)}

        return posteriors, training_metrics


class PyroEstimatorMAP(PyroEstimator):
    """Pyro Estimator for MAP Posteriors. DEPRECATED."""
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def fit(self, model_name, model_param_names, data_input, fitter=None, init_values=None):
        verbose = self.verbose
        message = self.message
        learning_rate = self.learning_rate
        seed = self.seed
        num_steps = self.num_steps
        learning_rate_total_decay = self.learning_rate_total_decay

        pyro.set_rng_seed(seed)
        if fitter is None:
            fitter = get_pyro_model(model_name)  # abstract
        model = fitter(data_input)  # concrete

        # Perform MAP inference using an AutoDelta guide.
        pyro.clear_param_store()
        guide = AutoDelta(model)
        optim = ClippedAdam({
            "lr": learning_rate,
            "lrd": learning_rate_total_decay ** (1 / num_steps),
            "betas": (0.5, 0.8)
        })
        elbo = Trace_ELBO()
        loss_elbo = list()
        svi = SVI(model, guide, optim, elbo)
        for step in range(num_steps):
            loss = svi.step()
            loss_elbo.append(loss)
            if verbose and step % message == 0:
                print("step {: >4d} loss = {:0.5g}".format(step, loss))

        # Extract point estimates.
        values = guide()
        values.update(pyro.poutine.condition(model, values)())

        # Convert from torch.Tensors to numpy.ndarrays.
        extract = {
            name: value.detach().numpy()
            for name, value in values.items()
        }

        # make sure that model param names are a subset of stan extract keys
        invalid_model_param = set(model_param_names) - set(list(extract.keys()))
        if invalid_model_param:
            raise EstimatorException("Pyro model definition does not contain required parameters")

        # `stan.optimizing` automatically returns all defined parameters
        # filter out unnecessary keys
        posteriors = {param: extract[param] for param in model_param_names}
        training_metrics = {'loss_elbo': np.array(loss_elbo)}

        return posteriors, training_metrics



import seaborn as sns
from matplotlib import pyplot as plt
import pandas as pd


from orbit.constants import palette

pd.options.mode.chained_assignment = None


def ts_heatmap(df, date_col, value_col, fig_width=10, fig_height=6, normalization=False,
               path=None, palette='Blues'):
    """this function takes a time series dataframe and plot a time series heatmap with month on the y axis and
    year on the x axis
    Parameters
    -----------
    df : pandas data frame
        input df
    date_col : str
        the name of the date column
    value_col : str
        the name of the value
    fig_width : int, optional
        adjust width of the chart
    fig_height : int, optional,
        adjust height of the chart
    normalization : bool, optional
        normalize using mean and std
    path : str
        path to save the figure
    palette : str, optional
        color palette

    Returns
    -------
    one time series heatmap chart
    """
    df = df[[date_col, value_col]]
    df[date_col] = pd.to_datetime(df[date_col])
    df['year'] = df[date_col].dt.year
    df['month'] = df[date_col].dt.month
    df_pivot = df.pivot_table(index='month', columns='year', values=value_col).sort_index(ascending=False)
    if normalization:
        df_pivot = (df_pivot - df_pivot.mean()) / df_pivot.std()
    fig, ax = plt.subplots(figsize=(fig_width, fig_height))
    ax = sns.heatmap(df_pivot, cmap=palette)

    if normalization:
        ax.set_title(f'{value_col} Time Series Heatmap Normalized Index')

    ax.set_title(f'{value_col} (Mean) Time Series Heatmap')

    if path:
        fig.savefig(path)

    return ax


def correlation_heatmap(df, var_list, fig_width=10, fig_height=6,
                        path=None, fmt='.1g', palette='Blues'):
    """This function takes a list of variables and return a heatmap of pairwise correlation. The columns with
    all zero values will not be plotted.
    Parameters
    -----------
    df : pandas data frame
        input df
    var_list : list
        list of the variable names
    fig_width : int, optional
        adjust width of the chart
    fig_height : int, optional,
        adjust height of the chart
    path : str
        path to save the figure
    palette : str, optional
        color palette

    Returns
    -------
    one correlation heatmap chart
    """
    # filter out all zero columns
    non_zero_varlist = [i for i in var_list if i not in df.columns[(df == 0).all()]]
    df = df[non_zero_varlist].corr()
    fig, ax = plt.subplots(figsize=(fig_width, fig_height))
    ax = sns.heatmap(df, cmap=palette, annot=True, fmt=fmt)
    ax.set_xticklabels(
        ax.get_xticklabels(),
        rotation=45,
        horizontalalignment='right'
    )
    ax.set_yticklabels(
        ax.get_yticklabels(),
        rotation=0,
        horizontalalignment='right'
    )
    ax.set_title('Correlation Heatmap')

    if path:
        fig.savefig(path)

    return ax


# def year_over_year_ts_plot(df, date_col, outcome, event_df, event_col,
#                            fig_width=30, fig_height=12, path=None):
#     """This chart plots outcome time series year over year against events to understand seasonality and trend, and how
#     events impact the pattern of outcome. The chart contains two subplots: 1) year over year time series plot of
#     Parameters
#     -----------
#     df : pandas data frame
#         input df
#     date_col : str
#         the name of the date column
#     outcome : str
#         the name of variable to be inspected
#     event_df : pandas data frame
#         event input data
#     event_col : str
#         name of the event column
#     fig_width : int, optional
#         adjust width of the chart
#     fig_height : int, optional,
#         adjust height of the chart
#     path : str
#         path to save the figure

#     Returns
#     -------
#     one chart with two subplots
#     """

#     df[date_col] = pd.to_datetime(df[date_col])
#     df['year'] = df[date_col].dt.year
#     df['month'] = df[date_col].dt.month
#     df['day'] = df[date_col].dt.day
#     df['day_of_month'] = df.month.astype(str) + '_' + df.day.astype(str)

#     event_df[date_col] = pd.to_datetime(event_df[date_col])
#     event_df['year'] = event_df[date_col].dt.year
#     event_df['month'] = event_df[date_col].dt.month
#     event_df['day'] = event_df[date_col].dt.day
#     event_df['day_of_month'] = event_df.month.astype(str) + '_' + event_df.day.astype(str)
#     event_df = event_df[event_df[date_col] <= df[date_col].max()]
#     color = iter(cm.rainbow(np.linspace(0, 1, event_df[event_col].nunique())))

#     fig, ax = plt.subplots(2, 1, figsize=(fig_width, fig_height))
#     # the first plot year over year time series vs events
#     for i in df.year.unique():
#         sns.lineplot(data=df[df['year'] == i], x='day_of_month', y=outcome, ax=ax[0], label=i)
#     ax[0].set_xticks(np.arange(0, 366, 14))
#     ax[0].title.set_text(f'{outcome} Year Over Year Plot')

#     h_list = event_df[event_col].unique()
#     for h in h_list:
#         h_df = event_df[event_df[event_col] == h]
#         c = next(color)
#         for i in h_df.day_of_month.unique():
#             ax[0].axvline(x=i, linestyle='--', color=c, label=h)
#         for d in h_df[date_col].unique():
#             ax[1].axvline(x=d, linestyle='--', color=c, label=h)
#     handles, labels = ax[0].get_legend_handles_labels()
#     handle_list, label_list = [], []
#     for handle, label in zip(handles, labels):
#         if label not in label_list:
#             handle_list.append(handle)
#             label_list.append(label)
#     ax[0].legend(handle_list, label_list, loc='center left', bbox_to_anchor=(1, 0.5), fancybox=True, shadow=True)
#     # the second plot: outcome vs event
#     sns.lineplot(data=df, x=date_col, y=outcome, ax=ax[1])
#     ax[1].legend(handle_list, label_list, loc='center left', bbox_to_anchor=(1, 0.5), fancybox=True, shadow=True)
#     ax[1].title.set_text(f'{outcome} Time Series with Events Plot')

#     if path:
#         fig.savefig(path)

#     return ax


def dual_axis_ts_plot(df, var1, var2, date_col, fig_width=25, fig_height=6, path=None,
                      color1=palette.QualitativePalette.mid_blue.value,
                      color2=palette.QualitativePalette.orange.value):
    """This function plots two time series variables on two y axis. This is handy for comparison of two variables. The dual
    y axis will set on two different scales if the two variables are very different in terms of volume.
    Parameters
    -----------
    df : pandas data frame
        input df
    var1 : str
        name of the first variable
    var2 : str
        name of the second variable
    date_col : str
        date column name
    fig_width : int, optional
        adjust width of the chart
    fig_height : int, optional,
        adjust height of the chart
    path : str
        path to save the figure
    color1: str
        color for var1
    color2: str
        color for var2

    Returns
    -------
    one time series plot with dual y axis
    """
    df[date_col] = pd.to_datetime(df[date_col])
    fig, ax = plt.subplots(figsize=(fig_width, fig_height))
    sns.lineplot(data=df, x=date_col, y=var1, label=var1, color=color1)
    ax.set_ylabel(f'{var1}', color=color1)
    ax.tick_params(axis='y', labelcolor=color1)
    ax2 = ax.twinx()
    sns.lineplot(data=df, x=date_col, y=var2, ax=ax2, color=color2, label=var2)
    ax2.set_ylabel(f'{var2}', color=color2)
    ax2.tick_params(axis='y', labelcolor=color2)
    lines, labels = ax.get_legend_handles_labels()
    lines2, labels2 = ax2.get_legend_handles_labels()
    ax2.legend(lines + lines2, labels + labels2, loc=0)
    ax.set_title(f'{var1} vs {var2}')
    # ax.grid(False)
    ax2.grid(False)

    if path:
        fig.savefig(path)

    return ax


def wrap_plot_ts(df, date_col, var_list, col_wrap=3, height=2.5, aspect=2):
    """This function plots a panel of time series plots.
    Parameters
    -----------
    df : pandas data frame
        input df
    date_col : str
        date column name
    var_list : list
        list of variable names

    Returns
    -------
    seaborn facet grid axes object
    """
    non_zero_varlist = [i for i in var_list if i not in df.columns[(df == 0).all()]]
    df = df[non_zero_varlist]
    df_long = df.melt(id_vars=[date_col])
    ax = sns.relplot(x=date_col, y='value', col='variable',
                     height=height, aspect=aspect,
                     col_wrap=col_wrap, kind='line', data=df_long,
                     facet_kws={'sharey': False, 'sharex': False}
                     )
    ax.set_xticklabels(rotation=45)
    plt.tight_layout()

    return ax


# def weekly_trend_decomposition(df, var, date_col):
#     """This function presents variable weekly trend after removing weekly seasonality.
#     Parameters
#     -----------
#     df : pandas data frame
#         input df
#     var : str
#         name of the variable
#     date_col : str
#         the name of the date column

#     Returns
#     -------
#     three subplots with Trend, Residual and Weekly Seasonality
#     """
#     df_dt = deepcopy(df)
#     df_dt.index = pd.to_datetime(df_dt[date_col])
#     res_weely = seasonal_decompose(df_dt[[var]], period=7, model='multiplicative')
#     fig, (ax1, ax2, ax3) = plt.subplots(3, 1, figsize=(15, 8))
#     res_weely.trend.plot(ax=ax1)
#     res_weely.resid.plot(ax=ax2)
#     res_weely.seasonal.plot(ax=ax3)
#     ax1.set_title("Removing Weekly Seasonality")
#     ax1.set_ylabel("Trend")
#     ax2.set_ylabel("Residual")
#     ax3.set_ylabel("Weekly Seasonality")
